# -*- coding: utf-8 -*-

''' local test file for developing, with pdf2docx installed by
    `python setpy.py develop`
'''

import os
from pdf2docx.converter import Converter
from pdf2docx.common.utils import compare_layput

if __name__ == '__main__':

    script_path = os.path.abspath(__file__) # current script path
    output = os.path.join(os.path.dirname(script_path), 'samples')
    filename = 'demo-table'
    pdf_file = os.path.join(output, f'{filename}.pdf')
    docx_file = os.path.join(output, f'{filename}.docx')

    cv = Converter(pdf_file, docx_file, debug=True)

    # process page by page
    for page in cv[0:1]:
        # parse layout
        cv.parse(page).make_page()
        
        # # extract tables
        # tables = cv.extract_tables(page)
        # for table in tables:
        #     print(table)
    
    cv.close() # close pdf

    # convert the docx file back to pdf manually, e.g. docx2pdf.pdf,
    # and compare results by checking bbox of each word.
    # The comparison result is stored in pdf file, e.g. comparison.pdf
    docx_pdf_file = os.path.join(output, f'docx2pdf.pdf')
    output_file = os.path.join(output, f'comparison.pdf')
    if compare_layput(pdf_file, docx_pdf_file, output_file, threshold=0.7):
        print('Fully matched.')


# -*- coding: utf-8 -*-

'''
Testing the layouts between sample pdf and the converted docx.

NOTE: pdf2docx should be installed in advance.

The docx is created from parsed layout, so an equivalent but efficient way is to 
compare parsed layout with banchmark one.

The benchmark layout is also created from parsed layout, but proved to be able to
create docx with good enough quality.

Obsoleted method (real-time testing method):
  - convert sample pdf to docx with this module
  - convert docx back to pdf (`OfficeToPDF` or online `pylovepdf`, see more link)
  - compare layouts between sample pdf and comparing pdf

more link:
  - https://github.com/cognidox/OfficeToPDF/releases
  - https://github.com/AndyCyberSec/pylovepdf
'''

import os
import unittest
import json

from pdf2docx.converter import Converter
from pdf2docx.layout.Layout import Layout
from pdf2docx.text.TextSpan import TextSpan


script_path = os.path.abspath(__file__) # current script path

class TestUtility(unittest.TestCase):
    '''utilities related directly to the test case'''

    @property
    def test_dir(self):
        return os.path.dirname(script_path)

    @property
    def layout_dir(self):
        return os.path.join(self.test_dir, 'layouts')

    @property
    def sample_dir(self):
        return os.path.join(self.test_dir, 'samples')

    @property
    def output_dir(self):
        return os.path.join(self.test_dir, 'outputs')

    def init_test(self, filename):
        ''' Initialize parsed layout and benchmark layout.'''
        # restore sample layout
        layout_file = os.path.join(self.layout_dir, f'{filename}.json')
        with open(layout_file, 'r') as f:
            raw_dict = json.load(f)
        self.sample = Layout(raw_dict)

        # parsed layout: first page only
        pdf_file = os.path.join(self.sample_dir, f'{filename}.pdf')
        docx_file = os.path.join(self.output_dir, f'{filename}.docx')
        cv = Converter(pdf_file, docx_file)        
        cv.parse(cv[0]).make_page()
        self.test = cv.layout # type: Layout
        cv.close()

        return self


    def verify_layout(self, threshold=0.95):
        ''' Check layout between benchmark and parsed one.'''
        self._check_text_layout(threshold)
        self._check_image_layout(threshold)
        self._check_table_layout(threshold)


    def _check_table_layout(self, threshold):
        '''Check table layout.
             - table contents are covered by text layout checking
             - check table structure
        '''
        sample_tables = self.sample.blocks.table_blocks
        test_tables = self.test.blocks.table_blocks

        # count of tables
        m, n = len(sample_tables), len(test_tables)
        self.assertEqual(m, n, msg=f"\nThe count of parsed tables '{n}' is inconsistent with sample '{m}'")

        # check structures table by table
        for sample_table, test_table in zip(sample_tables, test_tables):
            for sample_row, test_row in zip(sample_table, test_table):
                for sample_cell, test_cell in zip(sample_row, test_row):
                    if not sample_cell: continue
                    matched, msg = sample_cell.compare(test_cell, threshold)
                    self.assertTrue(matched, msg=f'\n{msg}')


    def _check_image_layout(self, threshold):
        '''Check image layout: bbox and vertical spacing.'''
        sample_image_spans = self.sample.blocks.image_spans(level=1)
        test_image_spans = self.test.blocks.image_spans(level=1)

        # count of images
        m, n = len(sample_image_spans), len(test_image_spans)
        self.assertEqual(m, n, msg=f"\nThe count of image blocks {n} is inconsistent with sample {m}")

        # check each image
        for sample, test in zip(sample_image_spans, test_image_spans):
            matched, msg = sample.compare(test, threshold)
            self.assertTrue(matched, msg=f'\n{msg}')


    def _check_text_layout(self, threshold):
        ''' Compare text layout and format. 
             - text layout is determined by vertical spacing
             - text format is defined in style attribute of TextSpan
        '''
        sample_text_blocks = self.sample.blocks.text_blocks(level=1)
        test_text_blocks = self.test.blocks.text_blocks(level=1)

        # count of blocks
        m, n = len(sample_text_blocks), len(test_text_blocks)
        self.assertEqual(m, n, msg=f"\nThe count of text blocks {n} is inconsistent with sample {m}")
        
        # check layout of each block
        for sample, test in zip(sample_text_blocks, test_text_blocks):

            # text bbox and vertical spacing
            matched, msg = sample.compare(test, threshold)
            self.assertTrue(matched, msg=f'\n{msg}')

            # text style
            for sample_line, test_line in zip(sample.lines, test.lines):
                for sample_span, test_span in zip(sample_line.spans, test_line.spans):
                    if not isinstance(sample_span, TextSpan): continue
                    
                    # text
                    a, b = sample_span.text, test_span.text
                    self.assertEqual(a, b, msg=f"\nApplied text '{b}' is inconsistent with sample '{a}'")

                    # style
                    m, n = len(sample_span.style), len(test_span.style)
                    self.assertEqual(m, n, msg=f"\nThe count of applied text style {n} is inconsistent with sample {m}")

                    for sample_dict, test_dict in zip(sample_span.style, test_span.style):
                        a, b = sample_dict['type'], test_dict['type']
                        self.assertEqual(a, b, msg=f"\nApplied text style '{b}' is inconsistent with sample '{a}'")


class MainTest(TestUtility):
    '''Main text class.'''

    def setUp(self):
        # create output path if not exist
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

    def test_text_format(self):
        '''sample file focusing on text format, e.g. highlight, underline, strike-through.'''
        self.init_test('demo-text').verify_layout(threshold=0.95)

    def test_image(self):
        '''sample file focusing on inline-image.'''
        self.init_test('demo-image').verify_layout(threshold=0.95)

    def test_table_format(self):
        '''sample file focusing on table format, e.g. 
            - border and shading style
            - vertical cell
            - merged cell
            - text format in cell
        '''
        self.init_test('demo-table').verify_layout(threshold=0.95)

    def test_text_scaling(self):
        '''sample file focusing on font size.
            In this case, the font size is set precisely with character scaling.
        '''
        self.init_test('demo-text-scaling').verify_layout(threshold=0.95)

    def test_path_transformation(self):
        '''sample file focusing on path transformation.
            In this case, the (0,0) origin is out of the page.
        '''
        self.init_test('demo-path-transformation').verify_layout(threshold=0.95)


# -*- coding: utf-8 -*-

'''
A group of Rows objects in a table.

@created: 2020-08-15
@author: train8808@gmail.com
'''

from .Row import Row
from ..common.Collection import Collection


class Rows(Collection):
    '''A group of Rows.'''

    def from_dicts(self, raws:list):
        for raw in raws:
            row = Row(raw)
            self.append(row)
        return self



# -*- coding: utf-8 -*-

'''
A group of Line objects.

@created: 2020-07-24
@author: train8808@gmail.com
'''


from .Line import Line
from ..common import utils
from ..common.Collection import Collection


class Lines(Collection):
    '''Text line list.'''

    def from_dicts(self, raws:list):
        for raw in raws:
            line = Line(raw)
            self.append(line)
        return self

    @property
    def image_spans(self):
        '''Get all ImageSpan instances.'''
        spans = []
        for line in self._instances:
            spans.extend(line.image_spans)
        return spans


    def intersects(self, line:Line):
        ''' Whether intersection exists between any line and given line.'''
        is_image_line = bool(line.image_spans)
        for instance in self._instances:

            # for image line, no any intersection is allowed
            if instance.image_spans or is_image_line:
                if instance.bbox.intersects(line.bbox):
                    return True
            
            # otherwise, the overlap tolerance is larger
            elif utils.get_main_bbox(instance.bbox, line.bbox, threshold=0.5):
                return True
        
        return False

    
    def merge(self):
        ''' Merge lines aligned horizontally in a block. The main purposes:
            - remove overlapped lines, e.g. floating images
            - make inline image as a span in text line logically
        '''
        # skip if empty
        if not self._instances: return self
    
        # sort lines
        self.sort()

        # check each line
        lines = Lines([self._instances[0]])
        for line in self._instances:
            
            # skip if intersection exists
            if lines.intersects(line):
                continue

            # add line directly if not aligned horizontally with previous line
            if not line.horizontally_align_with(lines[-1]):
                lines.append(line)
                continue

            # if it exists x-distance obviously to previous line,
            # take it as a separate line as it is
            if abs(line.bbox.x0-lines[-1].bbox.x1) > utils.DM:
                lines.append(line)
                continue

            # now, this line will be append to previous line as a span
            lines[-1].add(list(line.spans))

        # update lines in block
        self.reset(list(lines))


    def split(self):
        ''' Split vertical lines.'''
        # get horizontally aligned lines group by group
        fun = lambda a,b: a.horizontally_align_with(b, factor=0.0)
        groups = self.group(fun)
        return groups


    def sort(self):
        ''' Sort lines considering text direction.
            Taking natural reading direction for example:
            reading order for rows, from left to right for lines in row.

            In the following example, A should come before B.
            ```
                             +-----------+
                +---------+  |           |
                |   A     |  |     B     |
                +---------+  +-----------+
            ```
            Steps:
              - sort lines in reading order, i.e. from top to bottom, from left to right.
              - group lines in row
              - sort lines in row: from left to right
        '''
        # sort in reading order
        self.sort_in_reading_order()

        # split lines in separate row
        lines_in_rows = [] # type: list[list[Line]]

        for line in self._instances:

            # add lines to a row group if not in same row with previous line
            if not lines_in_rows or not line.in_same_row(lines_in_rows[-1][-1]):
                lines_in_rows.append([line])
            
            # otherwise, append current row group
            else:
                lines_in_rows[-1].append(line)
        
        # sort lines in each row: consider text direction
        idx = 0 if self.is_horizontal else 3
        self._instances = []
        for row in lines_in_rows:
            row.sort(key=lambda line: line.bbox_raw[idx])
            self._instances.extend(row)


# -*- coding: utf-8 -*-

'''
A group of TextSpan and ImageSpan objects.

@created: 2020-07-23
@author: train8808@gmail.com
'''

from ..common.Collection import Collection
from .TextSpan import TextSpan
from .ImageSpan import ImageSpan

class Spans(Collection):
    '''Text span list.'''

    def from_dicts(self, raws:list):
        for raw_span in raws:
            span = ImageSpan(raw_span) if 'image' in raw_span else TextSpan(raw_span)
            self.append(span)
        return self

    @property
    def text_spans(self):
        spans = list(filter(
            lambda span: isinstance(span, TextSpan), self._instances
        ))
        return Spans(spans)

    @property
    def image_spans(self):
        spans = list(filter(
            lambda span: isinstance(span, ImageSpan), self._instances
        ))
        return Spans(spans)


# -*- coding: utf-8 -*-

'''
Layout objects based on PDF raw dict extracted with PyMuPDF.

@created: 2020-07-22
@author: train8808@gmail.com
---

The raw page content extracted with PyMuPDF, `page.getText('rawdict')` is described per link:
https://pymupdf.readthedocs.io/en/latest/textpage.html

In addition to the raw layout dict, some new features are also included, e.g.
    - page margin
    - rectangle shapes, for text format, annotations and table border/shading
    - new block in table type

{
    # raw dict
    ----------------------------
    "width" : w,
    "height": h,    
    "blocks": [{...}, {...}, ...],

    # introduced dict
    ----------------------------
    "margin": [left, right, top, bottom],
    "rects" : [{...}, {...}, ...]
}
'''


import json
from docx.shared import Pt
from docx.enum.section import WD_SECTION

from .Blocks import Blocks
from ..shape.Rectangles import Rectangles
from ..table.TablesConstructor import TablesConstructor
from ..common.base import PlotControl
from ..common.utils import debug_plot
from ..common.pdf import new_page_with_margin
from ..common.docx import reset_paragraph_format


class Layout:
    ''' Object representing the whole page, e.g. margins, blocks, shapes, spacing.'''

    def __init__(self, raw:dict) -> None:
        self.width = raw.get('width', 0.0)
        self.height = raw.get('height', 0.0)

        # initialize blocks
        self.blocks = Blocks().from_dicts(raw.get('blocks', []))

        # initialize rects: to add rectangles later
        self.rects = Rectangles()

        # table parser
        self._tables_constructor = TablesConstructor(self.blocks, self.rects)

        # page margin: to calculate after cleaning blocks
        self._margin = None

    @property
    def margin(self):
        return self._margin


    def store(self):
        return {
            'width': self.width,
            'height': self.height,
            'margin': self._margin,
            'blocks': self.blocks.store(),
            'rects': self.rects.store(),
        }


    def serialize(self, filename:str):
        '''Write layout to specified file.'''
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.store(), indent=4))


    def plot(self, doc, title:str, key:PlotControl=PlotControl.LAYOUT):
        '''Plot specified type of blocks layout with PyMuPDF.
            ---
            Args:
              - doc: fitz.Document object
        '''
        # get objects to plot
        #  - all blocks
        if key == PlotControl.LAYOUT: 
            objects = list(self.blocks)
        
        #  - explicit table structure only
        elif key == PlotControl.TABLE: 
            objects = self.blocks.explicit_table_blocks
        
        #  - implicit table structure only
        elif key == PlotControl.IMPLICIT_TABLE: 
            objects = self.blocks.implicit_table_blocks
        
        #  - rectangle shapes
        elif key == PlotControl.SHAPE: 
            objects = list(self.rects)

        else:
            objects = []

        # do nothing if no objects
        if not objects: return

        # insert a new page
        page = new_page_with_margin(doc, self.width, self.height, self.margin, title)

        # plot styled table but no text blocks in cell
        if key==PlotControl.TABLE: 
            for item in objects:
                item.plot(page, content=False, style=True)
        
        # plot non-styled table and no text blocks in cell
        elif key==PlotControl.IMPLICIT_TABLE: 
            for item in objects:
                item.plot(page, content=False, style=False)
        
        else:
            for item in objects:
                 item.plot(page) # default args for TableBlock.plot


    def parse(self, **kwargs):
        ''' Parse page layout.
            ---
            Args:
              - kwargs: dict for layout plotting
                    kwargs = {
                        'debug': bool,
                        'doc': fitz.Document object or None,
                        'filename': str
                    }
        '''

        # preprocessing, e.g. change block order, clean negative block
        self.clean(**kwargs)        
    
        # parse table blocks: 
        #  - table structure/format recognized from rectangles
        self.parse_explicit_tables(**kwargs)
        
        #  - cell contents extracted from text blocks
        self.parse_implicit_tables(**kwargs)

        # parse text format, e.g. highlight, underline
        self.parse_text_format(**kwargs)
        
        # paragraph / line spacing
        self.parse_vertical_spacing()


    def extract_tables(self):
        '''Extract content from explicit tables.'''
        # parsing tables
        self.clean().parse_explicit_tables()

        # check table
        tables = [] # type: list[ list[list[str]] ]
        for table_block in self.blocks.table_blocks:
            tables.append(table_block.text)

        return tables


    def make_page(self, doc):
        ''' Create page based on layout data. 

            To avoid incorrect page break from original document, a new page section
            is created for each page.

            Support general document style only:
              - writing mode: from left to right, top to bottom
              - text direction: horizontal

            The vertical postion of paragraph/table is defined by space_before or 
            space_after property of a paragraph.
        '''
        # new page section
        # a default section is created when initialize the document,
        # so we do not have to add section for the first time.
        if not doc.paragraphs:
            section = doc.sections[0]
        else:
            section = doc.add_section(WD_SECTION.NEW_PAGE)

        section.page_width  = Pt(self.width)
        section.page_height = Pt(self.height)

        # set page margin
        left,right,top,bottom = self.margin
        section.left_margin = Pt(left)
        section.right_margin = Pt(right)
        section.top_margin = Pt(top)
        section.bottom_margin = Pt(bottom)

        # add paragraph or table according to parsed block
        for block in self.blocks:           

            # make paragraphs
            if block.is_text_block():
                # new paragraph
                p = doc.add_paragraph()
                block.make_docx(p, self.margin)
            
            # make table
            elif block.is_table_block():

                # create dummy paragraph if table before space is set
                # line spacing: table before_space/2.0
                # before space: table before space / 2.0
                if block.before_space:
                    h = round(block.before_space/2.0, 1)
                    p = doc.add_paragraph()
                    pf = reset_paragraph_format(p)
                    pf.space_before = Pt(max(h, 0.0))
                    pf.space_after = Pt(0)
                    pf.line_spacing = Pt(h)

                # new table            
                table = doc.add_table(rows=block.num_rows, cols=block.num_cols)
                table.autofit = False
                table.allow_autofit  = False
                block.make_docx(table, self.margin)
                
        # NOTE: If a table is at the end of a page, a new paragraph will be automatically 
        # added by the rending engine, e.g. MS Word, which resulting in an unexpected
        # page break. The solution is to never put a table at the end of a page, so add
        # an empty paragraph and reset its format, particularly line spacing, when a table
        # is created.
        if len(self.blocks) and self.blocks[-1].is_table_block():
            p = doc.add_paragraph()
            reset_paragraph_format(p, Pt(1.0)) # a small line height: 1 Pt


    @debug_plot('Clean Blocks and Shapes', plot=True, category=PlotControl.LAYOUT)
    def clean(self, **kwargs):
        '''Clean blocks and rectangles, e.g. remove negative blocks, duplicated rects.'''
        clean_blocks = self.blocks.clean()
        clean_rects  = self.rects.clean()
        
        # calculate page margin based on clean layout
        self._margin = self.blocks.page_margin(self.width, self.height)

        return clean_blocks or clean_rects


    @debug_plot('Explicit Table Structure', plot=True, category=PlotControl.TABLE)
    def parse_explicit_tables(self, **kwargs) -> bool:
        '''parse table structure from rectangle shapes'''
        tables = self._tables_constructor.explicit_tables()
        return bool(tables)


    @debug_plot('Implicit Table Structure', plot=True, category=PlotControl.IMPLICIT_TABLE)
    def parse_implicit_tables(self, **kwargs):
        ''' Parse table structure based on the layout of text/image blocks.

            Since no cell borders exist in this case, there may be various probabilities of table structures. 
            Among which, we use the simplest one, i.e. 1-row and n-column, to make the docx look like pdf.

            Ensure no horizontally aligned blocks in each column, so that these blocks can be converted to
            paragraphs consequently in docx.
        '''
        # horizontal range of table
        left, right, *_ = self.margin
        X0, X1 = left, self.width - right

        tables = self._tables_constructor.implicit_tables(X0, X1)
        return bool(tables)


    @debug_plot('Parsed Text Blocks', plot=True, category=PlotControl.LAYOUT)
    def parse_text_format(self, **kwargs):
        '''Parse text format in both page and table context.'''
        return self.blocks.parse_text_format(self.rects)
 
 
    def parse_vertical_spacing(self):
        ''' Calculate external and internal vertical space for paragraph blocks under page context 
            or table context. It'll used as paragraph spacing and line spacing when creating paragraph.
        '''
        left, right, top, bottom = self.margin
        bbox = (left, top, self.width-right, self.height-bottom)
        self.blocks.parse_vertical_spacing(bbox)



# -*- coding: utf-8 -*-

'''
Object with a boundary box, e.g. Block, Line, Span.
'''

import copy
import fitz
from .base import IText
from .utils import get_main_bbox

class BBox(IText):
    '''Boundary box with attribute in fitz.Rect type.'''
    def __init__(self, raw:dict={}):
        bbox = raw.get('bbox', (0,0,0,0))
        self._bbox = tuple([round(x,1) for x in bbox])

    def __bool__(self):
        '''Real object when bbox is defined.'''
        return bool(self.bbox)
    
    @property
    def bbox_raw(self):
        '''top-left, bottom-right points of bbox, (x0, y0, x1, y1).'''
        return self._bbox

    @property
    def bbox(self):
        '''bbox in fitz.Rect type.'''
        return fitz.Rect(self._bbox) if self._bbox else fitz.rect()

    
    @property
    def is_valid(self):
        '''Ensure bbox can be shown in page, i.e. all coordinates are positive.'''
        return all(x>=0 for x in self.bbox_raw)
    
   
    def vertically_align_with(self, bbox, factor:float=0.0, text_direction:bool=True):
        ''' Check whether two boxes have enough intersection in vertical direction, i.e. perpendicular to reading direction.
            ---
            Args:
              - bbox: BBox to check with
              - factor: threshold of overlap ratio, the larger it is, the higher probability the two bbox-es are aligned.
              - text_direction: consider text direction or not. True by default, from left to right if False.

            ```
            +--------------+
            |              |
            +--------------+ 
                    L1
                    +-------------------+
                    |                   |
                    +-------------------+
                            L2
            ```
            
            An enough intersection is defined based on the minimum width of two boxes:
            ```
            L1+L2-L>factor*min(L1,L2)
            ```
        '''
        if not bbox: return False

        # text direction
        is_horizontal = self.is_horizontal if text_direction else True
        idx = 0 if is_horizontal else 1

        L1 = self.bbox_raw[idx+2]-self.bbox_raw[idx]
        L2 = bbox.bbox_raw[idx+2]-bbox.bbox_raw[idx]
        L = max(self.bbox_raw[idx+2], bbox.bbox_raw[idx+2]) - min(self.bbox_raw[idx], bbox.bbox_raw[idx])

        return L1+L2-L>=factor*max(L1,L2)


    def horizontally_align_with(self, bbox, factor:float=0.0, text_direction:bool=True):
        ''' Check whether two boxes have enough intersection in horizontal direction, i.e. along the reading direction.
            ---
            Args:
              - bbox: BBox to check with
              - factor: threshold of overlap ratio, the larger it is, the higher probability the two bbox-es are aligned.
              - text_direction: consider text direction or not. True by default, from left to right if False.

            ```
            +--------------+
            |              | L1  +--------------------+
            +--------------+     |                    | L2
                                 +--------------------+
            ```
            
            An enough intersection is defined based on the minimum width of two boxes:
            ```
            L1+L2-L>factor*min(L1,L2)
            ```
        '''
        if not bbox: return False

        # text direction
        is_horizontal = self.is_horizontal if text_direction else True
        idx = 1 if is_horizontal else 0
        
        L1 = self.bbox_raw[idx+2]-self.bbox_raw[idx]
        L2 = bbox.bbox_raw[idx+2]-bbox.bbox_raw[idx]
        L = max(self.bbox_raw[idx+2], bbox.bbox_raw[idx+2]) - min(self.bbox_raw[idx], bbox.bbox_raw[idx])

        return L1+L2-L>=factor*max(L1,L2)


    def copy(self):
        '''make a deep copy.'''
        return copy.deepcopy(self)

    def update(self, rect):
        '''Update current bbox to specified `rect`.
            ---
            Args:
              - rect: fitz.rect or raw bbox like (x0, y0, x1, y1)
        '''
        fitz_rect = fitz.Rect(rect)
        bbox = (fitz_rect.x0, fitz_rect.y0, fitz_rect.x1, fitz_rect.y1)
        self._bbox = tuple([round(x,1) for x in bbox])
        return self

    def union(self, bbox):
        '''Update current bbox to the union with specified `rect`.
            ---
            Args:
              - bbox: BBox, the target to get union
        '''
        return self.update(self.bbox | bbox.bbox)


    def compare(self, bbox, threshold=0.9):
        '''Whether has same type and bbox.'''
        if not isinstance(bbox, self.__class__):
            return False, f'Inconsistent type: {self.__class__.__name__} v.s. {bbox.__class__.__name__}'
        
        if self.bbox_raw!=bbox.bbox_raw and not get_main_bbox(self.bbox, bbox.bbox, threshold):
            return False, f'Inconsistent bbox: {self.bbox_raw} v.s. {bbox.bbox_raw}'
        
        return True, ''

    def store(self):
        '''Store in json format.'''
        return { 'bbox': self._bbox }


# -*- coding: utf-8 -*-

'''
Image Span based on same raw data structure with image block.

@created: 2020-07-22
@author: train8808@gmail.com
'''

from ..common import docx
from .Image import Image


class ImageSpan(Image):
    '''Image span.'''
    def __init__(self, raw:dict={}):
        super(ImageSpan, self).__init__(raw)


    def store(self):
        return super(ImageSpan, self).store_image()


    def plot(self, page, color:tuple):
        '''Plot image bbox with diagonal lines.
            ---
            Args: 
              - page: fitz.Page object
        '''
        x0, y0, x1, y1 = self.bbox_raw
        page.drawLine((x0, y0), (x1, y1), color=color, width=1)
        page.drawLine((x0, y1), (x1, y0), color=color, width=1)
        page.drawRect(self.bbox, color=color, fill=None, overlay=False)


    def intersects(self, rect):
        '''Create new ImageSpan object with image contained in given bbox.
            ---
            Args:
              - rect: fitz.Rect, target bbox
        '''
        # add image span directly only if fully contained in bbox
        if rect.contains(self.bbox):
            return self.copy()
        
        # otherwise, ignore image
        else:
            return ImageSpan()


    def make_docx(self, paragraph):
        '''Add image span to a docx paragraph.'''
        # add image
        docx.add_image(paragraph, self.image, self.bbox.x1-self.bbox.x0)


# -*- coding: utf-8 -*-

'''
Base classes.
'''

from enum import Enum


class BlockType(Enum):
    '''Block types.'''
    UNDEFINED = -1
    TEXT = 0
    IMAGE = 1
    EXPLICIT_TABLE = 2
    IMPLICIT_TABLE = 3


class RectType(Enum):
    ''' Rectangle type in context:
        - not defined   : -1
        - highlight     : 0
        - underline     : 1
        - strike-through: 2
        - table border  : 10
        - cell shading  : 11
    '''
    UNDEFINED = -1
    HIGHLIGHT = 0
    UNDERLINE = 1
    STRIKE = 2
    BORDER = 10
    SHADING = 11


class TextDirection(Enum):
    '''Block types.'''
    IGNORE     = -1
    LEFT_RIGHT = 0 # from left to right within a line, and lines go from top to bottom
    BOTTOM_TOP = 1 # from bottom to top within a line, and lines go from left to right


class PlotControl(Enum):
    ''' Control what to show when plotting blocks.
        - layout         : plot all blocks
        - table          : plot explicit table blocks only
        - implicit_table : plot implicit table blocks only
        - shape          : plot rectangle shapes
    '''
    LAYOUT = 0
    TABLE = 1
    IMPLICIT_TABLE = 2
    SHAPE = 3


class IText:
    '''Text related interface considering text direction.'''
    @property
    def text_direction(self):
        '''Default text direction: from left to right.'''
        return TextDirection.LEFT_RIGHT

    @property
    def is_horizontal(self):
        '''Check whether text direction is from left to right.'''
        return self.text_direction == TextDirection.LEFT_RIGHT

    @property
    def is_vertical(self):
        return self.text_direction == TextDirection.BOTTOM_TOP


# -*- coding: utf-8 -*-

'''
Text Line objects based on PDF raw dict extracted with PyMuPDF.
@created: 2020-07-22
@author: train8808@gmail.com
---

Data structure of line in text block:
{
    'bbox': (x0,y0,x1,y1),
    'wmode': m,
    'dir': [x,y],
    'spans': [ spans ]
}

https://pymupdf.readthedocs.io/en/latest/textpage.html
'''

from ..common.BBox import BBox
from ..common.base import TextDirection
from .Spans import Spans
from .ImageSpan import ImageSpan


class Line(BBox):
    '''Object representing a line in text block.'''
    def __init__(self, raw:dict={}) -> None:
        super(Line, self).__init__(raw)
        self.wmode = raw.get('wmode', 0) # writing mode
        self.dir = raw.get('dir', [1, 0]) # writing direction

        # collect spans
        self.spans = Spans(None, self).from_dicts(raw.get('spans', []))

    
    @property
    def text(self):
        '''Joining span text.'''
        spans_text = [span.text for span in self.spans]        
        return ''.join(spans_text)


    @property
    def image_spans(self):
        '''Get image spans in this Line.'''
        return list(filter(
            lambda span: isinstance(span, ImageSpan), self.spans
        ))

    @property
    def text_direction(self):
        if self.dir[0] == 1.0:
            return TextDirection.LEFT_RIGHT
        elif self.dir[1] == -1.0:
            return TextDirection.BOTTOM_TOP
        else:
            return TextDirection.IGNORE

    def store(self) -> dict:
        res = super().store()
        res.update({
            'wmode': self.wmode,
            'dir': self.dir,
            'spans': [
                span.store() for span in self.spans
            ]
        })

        return res


    def plot(self, page, color:int):
        '''Plot line border in red.
           ---
            Args: 
              - page: fitz.Page object
        '''
        page.drawRect(self.bbox, color=color, fill=None, overlay=False)


    def add(self, span_or_list):
        '''Add span list to current Line.
            ---
            Args:
              - span_or_list: a TextSpan or TextSpan list
        '''
        if isinstance(span_or_list, (list, tuple)):
            for span in span_or_list:
                self.add_span(span)
        else:
            self.add_span(span_or_list)


    def add_span(self, span:BBox):
        '''Add span to current Line.'''
        if isinstance(span, BBox):
            self.spans.append(span)


    def intersects(self, rect):
        '''Create new Line object with spans contained in given bbox.
            ---
            Args:
              - rect: fitz.Rect, target bbox
        '''
        # add line directly if fully contained in bbox
        if rect.contains(self.bbox):
            return self.copy()

        # further check spans in line
        # new line with same text attributes
        line = Line({
            'wmode': self.wmode,
            'dir': self.dir
        })
        for span in self.spans:
            contained_span = span.intersects(rect)
            line.add(contained_span)

        return line

    def in_same_row(self, line):
        ''' Check whether in same row/line with specified line. Note text direction.

            taking horizontal text as an example:
            - yes: the bottom edge of each box is lower than the centerline of the other one;
            - otherwise, not in same row.

            Note the difference with method `horizontally_align_with`. They may not in same line, though
            aligned horizontally.
        '''
        if not line or self.text_direction != line.text_direction:
            return False

        # normal reading direction by default
        idx = 1 if self.is_horizontal else 0

        c1 = (self.bbox_raw[idx] + self.bbox_raw[idx+2]) / 2.0
        c2 = (line.bbox_raw[idx] + line.bbox_raw[idx+2]) / 2.0

        # Note y direction under PyMuPDF context
        res = c1<=line.bbox_raw[idx+2] and c2<=self.bbox_raw[idx+2]
        return res
            


# -*- coding: utf-8 -*-

'''
Image object.
'''

import fitz
from ..common.BBox import BBox


class Image(BBox):
    '''Base image object.'''
    def __init__(self, raw:dict={}) -> None:
        super(Image, self).__init__(raw)
        self.ext = raw.get('ext', 'png')
        self.width = raw.get('width', 0.0)
        self.height = raw.get('height', 0.0)
        self._image = raw.get('image', b'')    # source image bytes

    @property
    def text(self):
        '''Return an image placeholder: "<image>".'''
        return '<image>'

    @property
    def image(self):
        '''Get the image bytes converted by PyMuPDF.

            The image bytes are existed already in the raw dict, but there is risk facing
            `docx.image.exceptions.UnrecognizedImageError` when recreate image with python-docx.
            As explained in link below:

            https://stackoverflow.com/questions/56405003/unrecognizedimageerror-image-insertion-error-python-docx

            The solution is to convert to bitmap with PyMuPDF in advance.
        '''
        return fitz.Pixmap(self._image).getImageData(output="png") # convert to png image


    def from_image(self, image):
        '''Update with image block.
            ---
            Args:
              - image: Image, target image block
        '''
        self.ext = image.ext
        self.width = image.width
        self.height = image.height
        self._image = image._image
        self.update(image.bbox)
        return self


    def store_image(self):
        res = super().store()
        res.update({
            'ext': self.ext,
            'width': self.width,
            'height': self.height,
            'image': '<image>' # drop real content to reduce size
        })

        return res


# -*- coding: utf-8 -*-

'''
Char object based on PDF raw dict extracted with PyMuPDF.

@created: 2020-07-22
@author: train8808@gmail.com
---

refer to: https://pymupdf.readthedocs.io/en/latest/textpage.html

raw dict for Char
    {
        'bbox'  : (x0, y0, x1, y1), 
        'c'     : str, 
        'origin': (x,y)
    }
'''


from ..common.BBox import BBox
from ..shape.Rectangle import Rectangle


class Char(BBox):
    '''Object representing a character.'''
    def __init__(self, raw:dict={}) -> None:
        super(Char, self).__init__(raw)
        self.c = raw.get('c', '')
        self.origin = raw.get('origin', None)       


    def contained_in_rect(self, rect:Rectangle, horizontal:bool=True):
        ''' Detect whether locates in a rect, or has an intersection 
            larger than half of the char bbox.
        '''
        # char in rect?
        if self.bbox in rect.bbox:
            return True

        # intersection?
        else:
            intsec = self.bbox & rect.bbox
            if horizontal:
                return intsec.width > 0.5*self.bbox.width
            else:
                return intsec.height > 0.5*self.bbox.height


    def store(self) -> dict:
        res = super().store()
        res.update({
            'c': self.c,
            'origin': self.origin
        })

        return res


# -*- coding: utf-8 -*-

'''
Image block objects based on PDF raw dict extracted with PyMuPDF.

@created: 2020-07-22
@author: train8808@gmail.com
---
https://pymupdf.readthedocs.io/en/latest/textpage.html

    {
        'type': 1,
        'bbox': (x0,y0,x1,y1),
        'ext': 'png',
        'width': w,
        'height': h,
        'image': b'',
        'colorspace': n,
        'xref': xref, 'yref': yref, 'bpc': bpc
    }

Note: the raw image block will be merged into text block: Text > Line > Span.

'''

from docx.shared import Pt
from .Line import Line
from .Image import Image
from .ImageSpan import ImageSpan
from .TextBlock import TextBlock
from ..common import utils
from ..common import docx
from ..common.Block import Block


class ImageBlock(Block, Image):
    '''Image block.'''
    def __init__(self, raw: dict):
        super(ImageBlock, self).__init__(raw)

        # set type
        self.set_image_block()


    def store(self):
        res = super(ImageBlock, self).store()
        res.update(
            super(ImageBlock, self).store_image()
        )
        return res


    def plot(self, page):
        '''Plot image bbox with diagonal lines.
            ---
            Args: 
              - page: fitz.Page object
        '''
        x0, y0, x1, y1 = self.bbox_raw
        color = utils.RGB_component_from_name('blue')

        page.drawLine((x0, y0), (x1, y1), color=color, width=1)
        page.drawLine((x0, y1), (x1, y0), color=color, width=1)
        page.drawRect(self.bbox, color=color, fill=None, overlay=False)


    def to_text_block(self):
        '''convert image block to text block: a span'''
        # image span
        span = ImageSpan().from_image(self)

        # add span to line
        image_line = Line()
        image_line.add(span)
        
        # insert line to block
        block = TextBlock()
        block.add(image_line)

        # set text block
        block.set_text_block()

        return block
        

    def parse_text_format(self, rects):
        '''parse text format with style represented by rectangles. Not necessary for image block.'''
        return False


    def make_docx(self, p, X0:float):
        ''' Create paragraph for an image block.
            ---
            Args:
              - p: docx paragraph instance
              - X0: left border of the paragraph
        '''
        # indent and space setting
        before_spacing = max(round(self.before_space, 1), 0.0)
        after_spacing = max(round(self.after_space, 1), 0.0)
        pf = docx.reset_paragraph_format(p)
        pf.space_before = Pt(before_spacing)
        pf.space_after = Pt(after_spacing)
        
        # restore default tabs
        pf.tab_stops.clear_all()

        # left indent implemented with tab
        pos = round(self.bbox.x0-X0, 2)
        pf.tab_stops.add_tab_stop(Pt(pos))
        docx.add_stop(p, Pt(pos), Pt(0.0))

        # create image
        docx.add_image(p, self.image, self.bbox.x1-self.bbox.x0)

        return p



# -*- coding: utf-8 -*-

'''
Text block objects based on PDF raw dict extracted with PyMuPDF.

@created: 2020-07-22
@author: train8808@gmail.com
---
https://pymupdf.readthedocs.io/en/latest/textpage.html

    {
        # raw dict
        # --------------------------------
        'type': 0,
        'bbox': (x0,y0,x1,y1),
        'lines': [ lines ]

        # introduced dict
        # --------------------------------
        'before_space': bs,
        'after_space': as,
        'line_space': ls
    }
'''

from docx.shared import Pt
from .Line import Line
from .Lines import Lines
from .ImageSpan import ImageSpan
from ..common.base import RectType, TextDirection
from ..common.Block import Block
from ..common import utils
from ..common import docx


class TextBlock(Block):
    '''Text block.'''
    def __init__(self, raw:dict={}) -> None:
        super(TextBlock, self).__init__(raw)

        # collect lines
        self.lines = Lines(None, self).from_dicts(raw.get('lines', []))

        # set type
        self.set_text_block()

    @property
    def text(self) -> str:
        '''Get text content in block, joning each line with `\n`.'''
        lines_text = [line.text for line in self.lines]
        return '\n'.join(lines_text)


    @property
    def sub_bboxes(self) -> list:
        '''bbox of sub-region, i.e. Line.'''
        return [line.bbox for line in self.lines]

    
    @property
    def text_direction(self):
        '''All lines contained in text block must have same text direction. Otherwise, set normal direction'''            
        res = set(line.text_direction for line in self.lines)
        # consider two text direction only:  left-right, bottom-top
        if TextDirection.IGNORE in res:
            return TextDirection.IGNORE
        elif len(res)==1:
            return list(res)[0]
        else:
            return TextDirection.LEFT_RIGHT


    def store(self) -> dict:
        res = super().store()
        res.update({
            'lines': self.lines.store()
        })
        return res


    def add(self, line:Line):
        '''Add line to TextBlock.'''
        if isinstance(line, Line):
            self.lines.append(line)


    def merge(self):
        '''Merge contained lines horizontally.'''
        self.lines.merge()


    def split(self):
        ''' Split contained lines vertically and create associated text blocks.'''
        blocks = [] # type: list[TextBlock]
        for lines in self.lines.split():
            text_block = TextBlock()
            text_block.lines.reset(list(lines))
            blocks.append(text_block)
        
        return blocks


    def plot(self, page):
        '''Plot block/line/span area, in PDF page.
           ---
            Args: 
              - page: fitz.Page object
        '''
        # block border in blue
        blue = utils.RGB_component_from_name('blue')   
        page.drawRect(self.bbox, color=blue, fill=None, overlay=False)

        # lines and spans
        for line in self.lines:
            # line border in red
            red = utils.RGB_component_from_name('red')
            line.plot(page, red)

            # span regions in random color
            for span in line.spans:
                c = utils.RGB_component_from_name('')                
                span.plot(page, c)


    def contains_discrete_lines(self, distance:float=25, threshold:int=2) -> bool:
        ''' Check whether lines in block are discrete: 
              - the count of lines with a distance larger than `distance` is greater then `threshold`.
              - ImageSpan exists
              - vertical text exists
        '''
        num = len(self.lines)
        if num==1: return False

        # check image spans
        if self.lines.image_spans:
            return True

        # check text direction
        if self.is_vertical:
            return True

        # check the count of discrete lines
        cnt = 1
        for i in range(num-1):
            line = self.lines[i]
            next_line = self.lines[i+1]

            if line.horizontally_align_with(next_line):
                # horizontally aligned but not in a same row -> discrete block
                if not line.in_same_row(next_line):
                    return True
                
                # otherwise, check the distance only
                else:
                    if abs(line.bbox.x1-next_line.bbox.x0) > distance:
                        cnt += 1

        return cnt >= threshold

    
    def parse_text_format(self, rects) -> bool:
        '''parse text format with style represented by rectangles.
            ---
            Args:
              - rects: Rectangles, potential styles applied on blocks
        '''
        flag = False

        # use each rectangle (a specific text format) to split line spans
        for rect in rects:

            # a same style rect applies on only one block
            if rect.type != RectType.UNDEFINED:
                continue

            # any intersection with current block?
            if not self.bbox.intersects(rect.bbox):
                # impossible to intersect with next rect if this block is behind current rect
                # NOTE: rects are sorted in reading order in advance
                if self.bbox.y1 < rect.bbox.y0: break
                continue

            # yes, then go further to lines in block            
            for line in self.lines:
                # any intersection in this line?
                intsec = rect.bbox & ( line.bbox + utils.DR )
                
                if not intsec: 
                    if rect.bbox.y1 < line.bbox.y0: break # lines must be sorted in advance
                    continue

                # yes, then try to split the spans in this line
                split_spans = []
                for span in line.spans: 
                    # include image span directly
                    if isinstance(span, ImageSpan):
                        split_spans.append(span)                   

                    # split text span with the format rectangle: span-intersection-span
                    else:
                        spans = span.split(rect, line.is_horizontal)
                        split_spans.extend(spans)
                        flag = True
                                                
                # update line spans                
                line.spans.reset(split_spans)

        return flag


    def parse_line_spacing(self):
        '''Calculate average line spacing.

            The layout of pdf text block: line-space-line-space-line, excepting space before first line, 
            i.e. space-line-space-line, when creating paragraph in docx. So, an average line height = space+line.

            Then, the height of first line can be adjusted by updating paragraph before-spacing.
        '''

        # check text direction
        idx = 1 if self.is_horizontal else 0

        ref_line = None
        count = 0

        for line in self.lines:
            # count of lines
            if not line.in_same_row(ref_line):
                count += 1

            # update reference line
            ref_line = line            
        
        bbox = self.lines[0].bbox_raw   # first line
        first_line_height = bbox[idx+2] - bbox[idx]
        block_height = self.bbox_raw[idx+2]-self.bbox_raw[idx]
        
        # average line spacing
        if count > 1:
            line_space = (block_height-first_line_height)/(count-1)
        else:
            line_space = block_height        
        self.line_space = line_space

        # since the line height setting in docx may affect the original bbox in pdf, 
        # it's necessary to update the before spacing:
        # taking bottom left corner of first line as the reference point                
        self.before_space += first_line_height - line_space

        # if before spacing is negative, set to zero and adjust calculated line spacing accordingly
        if self.before_space < 0:
            self.line_space += self.before_space / count
            self.before_space = 0.0


    def make_docx(self, p, bbox:tuple):
        ''' Create paragraph for a text block. Join line sets with TAB and set position according to bbox.
            ---
            Args:
              - p: docx paragraph instance
              - bbox: boundary box of paragraph

            Generally, a pdf block is a docx paragraph, with block|line as line in paragraph.
            But without the context, it's not able to recognize a block line as word wrap, or a 
            separate line instead. A rough rule used here:
            
            block line will be treated as separate line by default, except
              - (1) this line and next line are actually in the same line (y-position)
              - (2) if the rest space of this line can't accommodate even one span of next line, 
                    it's supposed to be normal word wrap.

            Refer to python-docx doc for details on text format:
            https://python-docx.readthedocs.io/en/latest/user/text.html            
        '''
        # check text direction
        # normal direction by default, taking left border as a reference
        # when from bottom to top, taking bottom border as a reference
        idx = 0 if self.is_horizontal else 3

        # indent and space setting
        before_spacing = max(round(self.before_space, 1), 0.0)
        after_spacing = max(round(self.after_space, 1), 0.0)
        pf = docx.reset_paragraph_format(p)
        pf.space_before = Pt(before_spacing)
        pf.space_after = Pt(after_spacing)
        
        # restore default tabs
        pf.tab_stops.clear_all()

        # set line spacing for text paragraph
        pf.line_spacing = Pt(round(self.line_space,1))
        current_pos = 0.0

        # set all tab stops
        all_pos = set([
            round(abs(line.bbox_raw[idx]-bbox[idx]), 2) for line in self.lines
            ])
        for pos in all_pos:
            if not pos: continue # ignore pos==0
            pf.tab_stops.add_tab_stop(Pt(pos))

        # add line by line
        for i, line in enumerate(self.lines):

            # left indent implemented with tab
            pos = round(abs(line.bbox_raw[idx]-bbox[idx]), 2)
            docx.add_stop(p, Pt(pos), Pt(current_pos))

            # add line
            for span in line.spans:
                # add content
                span.make_docx(p)

            # break line? new line by default
            line_break = True
            # no more lines after last line
            if line==self.lines[-1]: 
                line_break = False            
            
            # do not break line if they're indeed in same line
            elif line.in_same_row(self.lines[i+1]):
                line_break = False
            
            if line_break:
                p.add_run('\n')
                current_pos = 0
            else:
                current_pos = round(abs(line.bbox_raw[(idx+2)%4]-bbox[idx]), 2)

        return p



# -*- coding: utf-8 -*-

'''
Collection of Cell instances.

@created: 2020-08-15
@author: train8808@gmail.com
'''

from .Cell import Cell
from ..common.Collection import Collection


class Cells(Collection):
    '''A group of Cells.'''
    def from_dicts(self, raws:list):
        for raw in raws:
            cell = Cell(raw)
            self.append(cell)
        return self
    
    def append(self, cell:Cell):
        '''Append a cell and update bbox accordingly.
            Rewrite method of parent class, since allow empty cells, i.e. merged cells, being added.
        '''
        self._instances.append(cell)
        if not self._parent is None: # Note: `if self._parent` does not work here
            self._parent.union(cell)




# -*- coding: utf-8 -*-

'''
Object representing rectangles and lines, which is parsed from both raw streams and annotations of pdf.

@created: 2020-07-22
@author: train8808@gmail.com
---

The context meaning of rectangle shape may be:
    - strike through line of text
    - under line of text
    - highlight area of text
    - table border
    - cell shading

Rectangle data structure:
    {
        'type': int,
        'bbox': (x0, y0, x1, y1),
        'color': sRGB_value
    }
'''


from ..common.BBox import BBox
from ..common.base import RectType
from ..common.utils import RGB_component

class Rectangle(BBox):
    ''' Rectangle shapes. 

        NOTE: any paths in PDF are converted to rectangles for this module. For instances,
        a rectangle path with stroke only in pdf is converted to four rectangles representing each border.
    '''
    def __init__(self, raw:dict={}):
        super(Rectangle, self).__init__(raw)
        self._type = RectType.UNDEFINED # no type by default
        self.color = raw.get('color', 0)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, rect_type: RectType):
        self._type = rect_type

    def store(self):
        res = super().store()
        res.update({
            'type': self._type.value,
            'color': self.color
        })
        return res

    def plot(self, page):
        '''Plot rectangle shapes with PyMuPDF.
            ---
            Args:
              - page: fitz.Page object
        '''
        color = [c/255.0 for c in RGB_component(self.color)]
        page.drawRect(self.bbox, color=color, fill=color, width=0, overlay=False)


# -*- coding: utf-8 -*-

'''
Text Span object based on PDF raw dict extracted with PyMuPDF.

@created: 2020-07-22
@author: train8808@gmail.com
---

Refer to: https://pymupdf.readthedocs.io/en/latest/textpage.html

data structure for Span
    {
        # raw dict
        ---------------------------
        'bbox': (x0,y0,x1,y1),
        'color': sRGB
        'font': fontname,
        'size': fontzise,
        'flags': fontflags,
        'chars': [ chars ],

        # added dict
        ----------------------------
        'text': text,
        'style': [
            {
                'type': int,
                'color': int
            },
            ...
        ]
    }
'''

from docx.shared import Pt, RGBColor
from docx.oxml.ns import qn

from .Char import Char
from ..common.BBox import BBox
from ..common.base import RectType
from ..common import utils, docx
from ..shape.Rectangle import Rectangle


class TextSpan(BBox):
    '''Object representing text span.'''
    def __init__(self, raw:dict={}) -> None:
        super(TextSpan, self).__init__(raw)
        self.color = raw.get('color', 0)
        self._font = raw.get('font', None)
        self.size = raw.get('size', 12.0)
        self.flags = raw.get('flags', 0)
        self.chars = [ Char(c) for c in raw.get('chars', []) ] # type: list[Char]

        # introduced attributes
        # a list of dict: { 'type': int, 'color': int }
        self.style = raw.get('style', []) 

    @property
    def font(self):
        '''Parse raw font name, e.g. 
            - split with '+' and '-': BCDGEE+Calibri-Bold, BCDGEE+Calibri -> Calibri
            - split with upper case : ArialNarrow -> Arial Narrow, but exception: NSimSUN -> NSimSUN
        
            NSimSUN refers to Chinese font name `新宋体`, so consider a localization mapping.
        '''
        # process on '+' and '-'
        font_name = self._font.split('+')[-1]
        font_name = font_name.split('-')[0]

        # mapping font name
        key = font_name.replace(' ', '').replace('-', '').replace('_', '').upper() # normalize mapping key
        font_name = utils.DICT_FONTS.get(key, font_name)

        # split with upper case letters
        blank = ' '
        font_name = ''.join(f'{blank}{x}' if x.isupper() else x for x in font_name).strip(blank)

        return font_name


    @property
    def text(self):
        '''Joining chars in text span'''
        chars = [char.c for char in self.chars]        
        return ''.join(chars)


    def add(self, char:Char):
        '''Add char and update bbox accordingly.'''
        self.chars.append(char)
        self.union(char)


    def store(self) -> dict:
        res = super().store()
        res.update({
            'color': self.color,
            'font': self.font,
            'size': self.size,
            'flags': self.flags,
            'chars': [
                char.store() for char in self.chars
            ],
            'text': self.text,
            'style': self.style
        })
        return res


    def plot(self, page, color:tuple):
        '''Fill bbox with given color.
           ---
            Args: 
              - page: fitz.Page object
        '''
        page.drawRect(self.bbox, color=color, fill=color, width=0, overlay=False)


    def split(self, rect:Rectangle, horizontal:bool=True):
        '''Split span with the intersection: span-intersection-span.'''
        # any intersection in this span?
        intsec = rect.bbox & self.bbox

        # no, then add this span as it is
        if not intsec: return [self]

        # yes, then split spans:
        # - add new style to the intersection part
        # - keep the original style for the rest
        split_spans = [] # type: list[TextSpan]

        # expand the intersection area, e.g. for strike through line,
        # the intersection is a `line`, i.e. a rectangle with very small height,
        # so expand the height direction to span height
        if horizontal:
            intsec.y0 = self.bbox.y0
            intsec.y1 = self.bbox.y1
        else:
            intsec.x0 = self.bbox.x0
            intsec.x1 = self.bbox.x1

        # calculate chars in the format rectangle
        # combine an index with enumerate(), so the second element is the char
        f = lambda items: items[1].contained_in_rect(rect, horizontal)
        index_chars = list(filter(f, enumerate(self.chars)))

        # then we get target chars in a sequence
        pos = index_chars[0][0] if index_chars else -1 # start index -1 if nothing found
        length = len(index_chars)
        pos_end = max(pos+length, 0) # max() is used in case: pos=-1, length=0

        # split span with the intersection: span-intersection-span
        # 
        # left part if exists
        if pos > 0:
            if horizontal:
                bbox = (self.bbox.x0, self.bbox.y0, intsec.x0, self.bbox.y1)
            else:
                bbox = (self.bbox.x0, intsec.y1, self.bbox.x1, self.bbox.y1)
            split_span = self.copy().update(bbox)
            split_span.chars = self.chars[0:pos]
            split_spans.append(split_span)

        # middle intersection part if exists
        if length > 0:
            bbox = (intsec.x0, intsec.y0, intsec.x1, intsec.y1)
            split_span = self.copy().update(bbox)
            split_span.chars = self.chars[pos:pos_end]            
            split_span.parse_text_style(rect, horizontal)  # update style
            split_spans.append(split_span)

        # right part if exists
        if pos_end < len(self.chars):
            if horizontal:
                bbox = (intsec.x1, self.bbox.y0, self.bbox.x1, self.bbox.y1)
            else:
                bbox = (self.bbox.x0, self.bbox.y0, self.bbox.x1, intsec.y0)
            split_span = self.copy().update(bbox)
            split_span.chars = self.chars[pos_end:]
            split_spans.append(split_span)

        return split_spans


    def parse_text_style(self, rect: Rectangle, horizontal:bool=True):
        '''Parse text style based on the position to a span bbox.'''

        # consider text format type only
        if rect.type==RectType.BORDER or rect.type==RectType.SHADING:
            return False

        # considering text direction
        idx = 1 if horizontal else 0

        # recognize text format based on rect and the span it applying to
        # region height
        h_rect = rect.bbox_raw[idx+2] - rect.bbox_raw[idx]
        h_span = self.bbox_raw[idx+2] - self.bbox_raw[idx]

        # distance to span bottom border
        d = self.bbox_raw[idx+2] - rect.bbox_raw[idx]

        # highlight: both the rect height and overlap must be large enough
        if h_rect >= 0.5*h_span:
            # In general, highlight color isn't white
            if rect.color != utils.RGB_value((1,1,1)) and utils.get_main_bbox(self.bbox, rect.bbox, 0.75): 
                rect.type = RectType.HIGHLIGHT
    
        # near to bottom of span? yes, underline
        elif d <= 0.25*h_span:
            rect.type = RectType.UNDERLINE

        # near to center of span? yes, strike-through-line
        elif 0.35*h_span < d < 0.75*h_span:
            rect.type = RectType.STRIKE

        # unknown style
        else:
            rect.type = RectType.UNDEFINED

        # check rect type again
        if rect.type==RectType.UNDEFINED: return False

        style =  {
            'type': rect.type.value,
            'color': rect.color
        }
        self.style.append(style)

        return True


    def intersects(self, rect):
        '''Create new TextSpan object with chars contained in given bbox.
            ---
            Args:
              - rect: fitz.Rect, target bbox
        '''
        # add span directly if fully contained in bbox
        if rect.contains(self.bbox):
            return self.copy()

        # no intersection
        if not rect.intersects(self.bbox):
            return TextSpan()

        # furcher check chars in span
        span = self.copy()
        span.chars.clear()
        span.update((0.0,0.0,0.0,0.0))

        for char in self.chars:
            if utils.get_main_bbox(char.bbox, rect, 0.55): # contains at least a half part
                span.chars.append(char)
                span.union(char)

        return span


    def make_docx(self, paragraph):
        '''Add text span to a docx paragraph.'''
        # set text
        docx_span = paragraph.add_run(self.text)        

        # set style
        # https://python-docx.readthedocs.io/en/latest/api/text.html#docx.text.run.Font

        # basic font style
        # line['flags'] is an integer, encoding bool of font properties:
        # bit 0: superscripted (2^0)
        # bit 1: italic (2^1)
        # bit 2: serifed (2^2)
        # bit 3: monospaced (2^3)
        # bit 4: bold (2^4)
        docx_span.superscript = bool(self.flags & 2**0)
        docx_span.italic = bool(self.flags & 2**1)
        docx_span.bold = bool(self.flags & 2**4)

        # font name
        font_name = self.font
        docx_span.font.name = font_name
        docx_span._element.rPr.rFonts.set(qn('w:eastAsia'), font_name) # set font for chinese characters
        docx_span.font.color.rgb = RGBColor(*utils.RGB_component(self.color))

        # font size
        # NOTE: only x.0 and x.5 is accepted in docx, so set character scaling accordingly
        # if the font size doesn't meet this condition.
        font_size = round(self.size*2)/2.0
        docx_span.font.size = Pt(font_size)

        # adjust by set scaling
        scale = self.size / font_size
        if abs(scale-1.0)>=0.01:
            docx.set_char_scaling(docx_span, scale)
        
        # font style parsed from PDF rectangles: 
        # e.g. highlight, underline, strike-through-line
        for style in self.style:
            
            t = style['type']
            # Built-in method is provided to set highlight in python-docx, but supports only limited colors;
            # so, set character shading instead if out of highlight color scope
            if t==RectType.HIGHLIGHT.value:
                docx.set_char_shading(docx_span, style['color'])

            # underline set with built-in method `font.underline` has a same color with text.
            # so, try to set a different color with xml if necessary
            elif t==RectType.UNDERLINE.value:
                if self.color==style['color']:
                    docx_span.font.underline = True
                else:
                    docx.set_char_underline(docx_span, style['color'])
            
            # same color with text for strike line
            elif t==RectType.STRIKE.value:
                docx_span.font.strike = True

        


# -*- coding: utf-8 -*-

'''
Table Cell object.

@created: 2020-07-23
@author: train8808@gmail.com
'''

from docx.shared import Pt
from ..text.TextBlock import TextBlock
from ..common.BBox import BBox
from ..common.utils import RGB_component
from ..common import docx
from ..layout import Blocks # avoid conflict


class Cell(BBox):
    ''' Cell object.'''
    def __init__(self, raw:dict={}):
        if raw is None: raw = {}
        super(Cell, self).__init__(raw)
        self.bg_color = raw.get('bg_color', None) # type: int
        self.border_color = raw.get('border_color', None) # type: tuple [int]
        self.border_width = raw.get('border_width', None) # type: tuple [float]
        self.merged_cells = raw.get('merged_cells', (1,1)) # type: tuple [int]

        # collect blocks
        # NOTE: The cell bbox is determined first, and then find blocks contained in this bbox.
        # so, don't update cell bbox when appending blocks, i.e. set parent=None.
        self.blocks = Blocks.Blocks().from_dicts(raw.get('blocks', []))


    @property
    def text(self) -> str:
        '''Text contained in this cell.'''
        return '\n'.join([block.text for block in self.blocks]) if bool(self) else None

    
    def compare(self, cell, threshold:float=0.9):
        '''whether has same structure with given Cell.
            ---
            Args:
              - cell: Cell instance to compare
              - threshold: two bboxes are considered same if the overlap area exceeds threshold.
        '''
        res, msg = super().compare(cell, threshold)
        if not res:
            return res, msg
        
        if self.bg_color != cell.bg_color:
            return False, f'Inconsistent background color @ Cell {self.bbox_raw}:\n{self.bg_color} v.s. {cell.bg_color}'

        if tuple(self.border_color) != tuple(cell.border_color):
            return False, f'Inconsistent border color @ Cell {self.bbox_raw}:\n{self.border_color} v.s. {cell.border_color}'

        if tuple(self.border_width) != tuple(cell.border_width):
            return False, f'Inconsistent border width @ Cell {self.bbox_raw}:\n{self.border_width} v.s. {cell.border_width}'

        if tuple(self.merged_cells) != tuple(cell.merged_cells):
            return False, f'Inconsistent count of merged cells @ Cell {self.bbox_raw}:\n{self.merged_cells} v.s. {cell.merged_cells}'

        return True, ''


    def store(self):
        if bool(self):
            res = super().store()
            res.update({
                'bg_color': self.bg_color,
                'border_color': self.border_color,
                'border_width': self.border_width,
                'merged_cells': self.merged_cells,
                'blocks': self.blocks.store()
            })
            return res
        else:
            return None


    def plot(self, page, content:bool=True, style:bool=True, color:tuple=None):
        '''Plot cell.
            ---
            Args:
              - page: fitz.Page object
              - content: plot text blocks if True
              - style: plot cell style if True, e.g. border width, shading; otherwise draw table border only
              - color: table border color when style=False              
        '''        
        # plot cell style
        if style:
            # border color and width
            bc = [x/255.0 for x in RGB_component(self.border_color[0])]
            w = self.border_width[0]

            # shading color
            if self.bg_color != None:
                sc = [x/255.0 for x in RGB_component(self.bg_color)] 
            else:
                sc = None
            page.drawRect(self.bbox, color=bc, fill=sc, width=w, overlay=False)
        
        # or just cell borders for illustration
        else:
            page.drawRect(self.bbox, color=color, fill=None, width=1, overlay=False)

        # plot blocks contained in cell
        if content:
            for block in self.blocks:
                block.plot(page)


    def add(self, block):
        ''' Add block to this cell. 
            ---
            Arg:
              - block: Block type

            Note: If the block is partly contained in a cell, it must deep into line -> span -> char.
        '''
        if not block.is_text_block():
            return

        # add block directly if fully contained in cell
        if self.bbox.contains(block.bbox):
            self.blocks.append(block)
            return
        
        # add nothing if no intersection
        if not self.bbox.intersects(block.bbox):
            return

        # otherwise, further check lines in block
        split_block = TextBlock()
        for line in block.lines:
            L = line.intersects(self.bbox)
            split_block.add(L)

        self.blocks.append(split_block)


    def make_docx(self, table, indexes, border_style:bool):
        '''Create docx table.
            ---
            Args:
              - table: docx table instance
              - indexes: (i, j), row and column indexes
              - border_style: whether set border style
        '''
        # ignore merged cells
        if not bool(self): return
        
        # set cell style
        # no borders for implicit table
        self._set_style(table, indexes, border_style)

        # clear cell margin
        # NOTE: the start position of a table is based on text in cell, rather than left border of table. 
        # They're almost aligned if left-margin of cell is zero.
        docx_cell = table.cell(*indexes)
        docx.set_cell_margins(docx_cell, start=0, end=0)

        # set vertical direction if contained text blocks are in vertical direction
        if self.blocks.is_vertical:
            docx.set_vertical_cell_direction(docx_cell)

        # insert text            
        first = True
        for block in self.blocks:
            if first:
                p = docx_cell.paragraphs[0]
                first = False
            else:
                p = docx_cell.add_paragraph()
            block.make_docx(p, self.bbox_raw)


    def _set_style(self, table, indexes, border_style=True):
        ''' Set python-docx cell style, e.g. border, shading, width, row height, 
            based on cell block parsed from PDF.
            ---
            Args:
              - table: python-docx table object
              - indexes: (i, j) index of current cell in table
              - border_style: set border style or not
        '''
        i, j = indexes
        docx_cell = table.cell(i, j)
        n_row, n_col = self.merged_cells

        # ---------------------
        # border style
        # ---------------------
        # NOTE: border width is specified in eighths of a point, with a minimum value of 
        # two (1/4 of a point) and a maximum value of 96 (twelve points)
        if border_style:
            keys = ('top', 'end', 'bottom', 'start')
            kwargs = {}
            for k, w, c in zip(keys, self.border_width, self.border_color):
                hex_c = f'#{hex(c)[2:].zfill(6)}'
                kwargs[k] = {
                    'sz': 8*w, 'val': 'single', 'color': hex_c.upper()
                }
            # merged cells are assumed to have same borders with the main cell        
            for m in range(i, i+n_row):
                for n in range(j, j+n_col):
                    docx.set_cell_border(table.cell(m, n), **kwargs)

        # ---------------------
        # merge cells
        # ---------------------        
        if n_row*n_col!=1:
            _cell = table.cell(i+n_row-1, j+n_col-1)
            docx_cell.merge(_cell)

        # ---------------------
        # cell width (cell height is set by row height)
        # ---------------------
        # experience: width of merged cells may change if not setting width for merged cells
        x0, y0, x1, y1 = self.bbox_raw
        docx_cell.width = Pt(x1-x0)

        # ---------------------
        # cell bg-color
        # ---------------------
        if self.bg_color!=None:
            docx.set_cell_shading(docx_cell, self.bg_color)


# -*- coding: utf-8 -*-

import random

import fitz
from fitz.utils import getColorList, getColorInfoList
from .base import PlotControl

# border margin
DM = 1.0

# inch to point
ITP = 72.0

# tolerant rectangle area
DR = fitz.Rect(-DM, -DM, DM, DM) / 2.0

# font name mapping
# special process on the key:
# - upper case
# - delete blanks, '-', '_'
DICT_FONTS = {
    'SIMSUN': '宋体', 
    'SIMHEI': '黑体', 
    'MICROSOFTYAHEI': '微软雅黑', 
    'MICROSOFTJHENGHEI': '微软正黑体', 
    'NSIMSUN': '新宋体', 
    'PMINGLIU': '新细明体', 
    'MINGLIU': '细明体', 
    'DFKAISB': '标楷体', # 'DFKAI-SB'
    'FANGSONG': '仿宋', 
    'KAITI': '楷体', 
    'FANGSONGGB2312': '仿宋_GB2312', # FANGSONG_GB2312
    'KAITIGB2312': '楷体_GB2312',   # KAITI_GB2312
    'LISU': '隶书',
    'YOUYUAN': '幼圆',
    'STXIHEI': '华文细黑',
    'STKAITI': '华文楷体',
    'STSONG': '华文宋体',
    'STZHONGSONG': '华文中宋',
    'STFANGSONG': '华文仿宋',
    'FZSHUTI': '方正舒体',
    'FZYAOTI': '方正姚体',
    'STCAIYUN': '华文彩云',
    'STHUPO': '华文琥珀',
    'STLITI': '华文隶书',
    'STXINGKAI': '华文行楷',
    'STXINWEI': '华文新魏'
}


def is_number(str_number):
    try:
        float(str_number)
    except:
        return False
    else:
        return True

def RGB_component_from_name(name:str=''):
    '''Get a named RGB color (or random color) from fitz predefined colors, e.g. 'red' -> (1.0,0.0,0.0).'''
    # get color index
    if name and name.upper() in getColorList():
        pos = getColorList().index(name.upper())
    else:
        pos = random.randint(0, len(getColorList())-1)
        
    c = getColorInfoList()[pos]
    return (c[1] / 255.0, c[2] / 255.0, c[3] / 255.0)


def RGB_component(srgb:int):
    '''srgb value to R,G,B components, e.g. 16711680 -> (255, 0, 0)'''
    # decimal to hex: 0x...
    s = hex(srgb)[2:].zfill(6)
    return [int(s[i:i+2], 16) for i in [0, 2, 4]]


def RGB_value(rgb:list):
    '''RGB components to decimal value, e.g. (1,0,0) -> 16711680'''
    res = 0
    for (i,x) in enumerate(rgb):
        res += int(x*(16**2-1)) * 16**(4-2*i)
    return int(res)


def CMYK_to_RGB(c:float, m:float, y:float, k:float, cmyk_scale:float=100):
    ''' CMYK components to GRB value.'''
    r = (1.0 - c / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    g = (1.0 - m / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    b = (1.0 - y / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    res = RGB_value((r, g, b)) # type: int
    return res


def get_main_bbox(bbox_1:fitz.Rect, bbox_2:fitz.Rect, threshold:float=0.95):
    ''' If the intersection of bbox_1 and bbox_2 exceeds the threshold, return the union of
        these two bbox-es; else return None.
    '''
    # areas
    b = bbox_1 & bbox_2
    a1, a2, a = bbox_1.getArea(), bbox_2.getArea(), b.getArea()

    # no intersection
    if not b: return fitz.Rect()

    # Note: if bbox_1 and bbox_2 intersects with only an edge, b is not empty but b.getArea()=0
    # so give a small value when they're intersected but the area is zero
    factor = a/min(a1,a2) if a else 1e-6
    if factor >= threshold:
        return bbox_1 | bbox_2
    else:
        return fitz.Rect()


def expand_centerline(start: list, end: list, width:float=2.0):
    ''' convert centerline to rectangle shape.
        centerline is represented with start/end points: (x0, y0), (x1, y1).
    '''
    h = width / 2.0
    x0, y0 = start
    x1, y1 = end

    # consider horizontal or vertical line only
    if x0==x1 or y0==y1:
        res = (x0-h, y0-h, x1+h, y1+h)
    else:
        res = None

    return res


def debug_plot(title:str, plot:bool=True, category:PlotControl=PlotControl.LAYOUT):
    ''' Plot layout / shapes for debug mode when the following conditions are all satisfied:
          - plot=True
          - layout has been changed: the return value of `func` is True
          - debug mode: kwargs['debug']=True
          - the pdf file to plot layout exists: kwargs['doc'] is not None        
        ---        
        Args:
          - title: page title
          - plot: plot layout/shape if true
          - category: PlotControl, what to plot
    '''
    def wrapper(func):
        def inner(*args, **kwargs):
            # execute function
            res = func(*args, **kwargs)

            # check if plot layout
            debug = kwargs.get('debug', False)
            doc = kwargs.get('doc', None)
            layout = args[0] # assert Layout object
            if plot and res and debug and doc is not None:                
                layout.plot(doc, title, category)
            return layout
        return inner
    return wrapper


def compare_layput(filename_source, filename_target, filename_output, threshold=0.7):
    ''' Compare layout of two pdf files:
        It's difficult to have an exactly same layout of blocks, but ensure they
        look like each other. So, with `extractWORDS()`, all words with bbox 
        information are compared.

        ```
        (x0, y0, x1, y1, "word", block_no, line_no, word_no)
        ```
    '''
    # fitz document
    source = fitz.open(filename_source) # type: fitz.Document
    target = fitz.open(filename_target) # type: fitz.Document

    # check count of pages
    # --------------------------
    if len(source) != len(target):
        msg='Page count is inconsistent with source file.'
        print(msg)
        return False
    
    flag = True
    errs = []
    for i, (source_page, target_page) in enumerate(zip(source, target)):

        # check position of each word
        # ---------------------------
        source_words = source_page.getText('words')
        target_words = target_page.getText('words')

        # sort by word
        source_words.sort(key=lambda item: (item[4], round(item[1],1), round(item[0],1)))
        target_words.sort(key=lambda item: (item[4], round(item[1],1), round(item[0],1)))

        if len(source_words) != len(target_words):
            msg='Words count is inconsistent with source file.'
            print(msg)
            return False

        # check each word and bbox
        for sample, test in zip(source_words, target_words):
            source_rect, target_rect = fitz.Rect(sample[0:4]), fitz.Rect(test[0:4])

            # draw bbox based on source layout
            source_page.drawRect(source_rect, color=(1,1,0), overlay=True) # source position
            source_page.drawRect(target_rect, color=(1,0,0), overlay=True) # current position

            # check bbox word by word: ignore small bbox, e.g. single letter bbox
            if not get_main_bbox(source_rect, target_rect, threshold):
                flag = False
                errs.append((f'{sample[4]} ===> {test[4]}', target_rect, source_rect))
        
    # save and close
    source.save(filename_output)
    target.close()
    source.close()

    # outputs
    for word, target_rect, source_rect in errs:
        print(f'Word "{word}": \nsample bbox: {source_rect}\ncurrent bbox: {target_rect}\n')

    return flag


# -*- coding: utf-8 -*-

'''
docx operation methods based on python-docx.
'''

from io import BytesIO
from docx.shared import Pt
from docx.oxml import OxmlElement, parse_xml
from docx.oxml.ns import qn, nsdecls
from docx.enum.text import WD_COLOR_INDEX
from docx.image.exceptions import UnrecognizedImageError
from docx.table import _Cell

from .utils import RGB_value, DM


def reset_paragraph_format(p, line_spacing:float=1.05):
    ''' Reset paragraph format, especially line spacing.
        ---
        Args:
        - p: docx paragraph instance
        
        Two kinds of line spacing, corresponding to the setting in MS Office Word:
        - line_spacing=1.05: single or multiple
        - line_spacing=Pt(1): exactly
    '''
    pf = p.paragraph_format
    pf.line_spacing = line_spacing # single by default
    pf.space_before = Pt(0)
    pf.space_after = Pt(0)
    pf.left_indent = Pt(0)
    pf.right_indent = Pt(0)
    pf.widow_control = True

    # do not adjust spacing between Chinese and Latin/number
    xml = r'<w:autoSpaceDE {} w:val="0"/>'.format(nsdecls('w'))
    p._p.get_or_add_pPr().insert(0, parse_xml(xml))

    xml = r'<w:autoSpaceDN {} w:val="0"/>'.format(nsdecls('w'))
    p._p.get_or_add_pPr().insert(0, parse_xml(xml))

    return pf


def add_stop(p, pos:float, current_pos:float):
    ''' Set horizontal position in current position with tab stop.
        ---
        Args: 
          - p: docx paragraph instance
          - pos: target position in Pt
          - current_pos: current position in Pt

        Note: multiple tab stops may exist in paragraph, 
              so tabs are added based on current position and target position.         
    '''
    # ignore small pos
    if pos < Pt(DM): return

    # add tab to reach target position
    for t in p.paragraph_format.tab_stops:
        if t.position < current_pos:
            continue
        elif t.position<pos or abs(t.position-pos)<=Pt(DM):
            p.add_run().add_tab()
        else:
            break


def add_image(p, byte_image, width):
    ''' Add image to paragraph.
        ---
        Args:
          - p: docx paragraph instance
          - byte_image: bytes for image source
          - width: image width
    '''
    docx_span = p.add_run()
    try:
        docx_span.add_picture(BytesIO(byte_image), width=Pt(width))
    except UnrecognizedImageError:
        print('Unrecognized Image.')
        return
    
    # exactly line spacing will destroy image display, so set single line spacing instead
    p.paragraph_format.line_spacing = 1.00


def set_char_scaling(p_run, scale:float=1.0):
    ''' Set character spacing: scaling. Font | Advanced | Character Spacing | Scaling.
        ---
        Args:
          - p_run: docx.text.run.Run, proxy object wrapping <w:r> element
          - scale: scaling factor
    '''
    p_run._r.get_or_add_rPr().insert(0, parse_xml(r'<w:w {} w:val="{}"/>'.format(nsdecls('w'), 100*scale)))


def set_char_shading(p_run, srgb:int):
    '''Set character shading color, in case the color is out of highlight color scope.
        ---
        Args:
        - p_run: docx.text.run.Run, proxy object wrapping <w:r> element
        - srgb: int, color value

        Read more:
        - http://officeopenxml.com/WPtextShading.php
    '''
    # try to set highlight first using python-docx built-in method
    # Here give 6/16 of the valid highlight colors
    color_map = {        
        RGB_value((1,0,0)): WD_COLOR_INDEX.RED,
        RGB_value((0,1,0)): WD_COLOR_INDEX.BRIGHT_GREEN,
        RGB_value((0,0,1)): WD_COLOR_INDEX.BLUE,
        RGB_value((1,1,0)): WD_COLOR_INDEX.YELLOW,
        RGB_value((1,0,1)): WD_COLOR_INDEX.PINK,
        RGB_value((0,1,1)): WD_COLOR_INDEX.TURQUOISE
    }
    if srgb in color_map:
        p_run.font.highlight_color = color_map[srgb]

    # set char shading
    else:
        c = hex(srgb)[2:].zfill(6)
        xml = r'<w:shd {} w:val="clear" w:color="auto" w:fill="{}"/>'.format(nsdecls('w'), c)
        p_run._r.get_or_add_rPr().insert(0, parse_xml(xml))


def set_char_underline(p_run, srgb:int):
    '''Set underline and color.
        ---
        Args:
        - p_run: docx.text.run.Run, proxy object wrapping <w:r> element
        - srgb: int, color value
    '''
    c = hex(srgb)[2:].zfill(6)
    xml = r'<w:u {} w:val="single" w:color="{}"/>'.format(nsdecls('w'), c)
    p_run._r.get_or_add_rPr().insert(0, parse_xml(xml))


def indent_table(table, indent:float):
    ''' indent table.
        ---
        Args:
          - table: docx table object
          - indent: indent value, the basic unit is 1/20 pt
    '''
    tbl_pr = table._element.xpath('w:tblPr')
    if tbl_pr:
        e = OxmlElement('w:tblInd')
        e.set(qn('w:w'), str(20*indent)) # basic unit 1/20 pt for openxml 
        e.set(qn('w:type'), 'dxa')
        tbl_pr[0].append(e)


def set_cell_margins(cell:_Cell, **kwargs):
    ''' Set cell margins. Provided values are in twentieths of a point (1/1440 of an inch).
        ---
        Args:
          - cell:  actual cell instance you want to modify
          - kwargs: a dict with keys: top, bottom, start, end
        
        Usage:
          - set_cell_margins(cell, top=50, start=50, bottom=50, end=50)
        
        Read more: 
          - https://blog.csdn.net/weixin_44312186/article/details/104944773
          - http://officeopenxml.com/WPtableCellMargins.php
    '''
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
    tcMar = OxmlElement('w:tcMar')
 
    for m in ['top', 'start', 'bottom', 'end']:
        if m in kwargs:
            node = OxmlElement("w:{}".format(m))
            node.set(qn('w:w'), str(kwargs.get(m)))
            node.set(qn('w:type'), 'dxa')
            tcMar.append(node)
 
    tcPr.append(tcMar)


def set_cell_shading(cell:_Cell, srgb:int):
    ''' set cell background-color.
        ---
        Args:
        - cell: actual cell instance you want to modify
        - srgb: RGB color value

        https://stackoverflow.com/questions/26752856/python-docx-set-table-cell-background-and-text-color
    '''
    c = hex(srgb)[2:].zfill(6)
    cell._tc.get_or_add_tcPr().append(parse_xml(r'<w:shd {} w:fill="{}"/>'.format(nsdecls('w'), c)))


def set_cell_border(cell:_Cell, **kwargs):
    '''
    Set cell`s border.
    
    Reference:
     - https://stackoverflow.com/questions/33069697/how-to-setup-cell-borders-with-python-docx
     - https://blog.csdn.net/weixin_44312186/article/details/104944110

    Usage:
    ```
        _set_cell_border(
            cell,
            top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
            bottom={"sz": 12, "color": "#00FF00", "val": "single"},
            start={"sz": 24, "val": "dashed", "shadow": "true"},
            end={"sz": 12, "val": "dashed"},
        )
    ```
    '''
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()

    # check for tag existence, if none found, then create one
    tcBorders = tcPr.first_child_found_in("w:tcBorders")
    if tcBorders is None:
        tcBorders = OxmlElement('w:tcBorders')
        tcPr.append(tcBorders)

    # list over all available tags
    for edge in ('start', 'top', 'end', 'bottom', 'insideH', 'insideV'):
        edge_data = kwargs.get(edge)
        if edge_data:
            tag = 'w:{}'.format(edge)

            # check for tag existence, if none found, then create one
            element = tcBorders.find(qn(tag))
            if element is None:
                element = OxmlElement(tag)
                tcBorders.append(element)

            # looks like order of attributes is important
            for key in ["sz", "val", "color", "space", "shadow"]:
                if key in edge_data:
                    element.set(qn('w:{}'.format(key)), str(edge_data[key]))


def set_vertical_cell_direction(cell:_Cell, direction:str='btLr'):
    '''Set vertical text direction for cell.
        ---
        Args:
          - direction: tbRl -- top to bottom, btLr -- bottom to top
        
        https://stackoverflow.com/questions/47738013/how-to-rotate-text-in-table-cells
    '''
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
    textDirection = OxmlElement('w:textDirection')
    textDirection.set(qn('w:val'), direction)  # btLr tbRl
    tcPr.append(textDirection)



# -*- coding: utf-8 -*-

'''
Row in a table.

@created: 2020-08-15
@author: train8808@gmail.com
'''

from docx.enum.table import WD_ROW_HEIGHT
from docx.shared import Pt

from .Cell import Cell
from .Cells import Cells
from ..common.BBox import BBox


class Row(BBox):
    '''Row in a table.'''
    def __init__(self, raw:dict={}):
        super(Row, self).__init__(raw)

        # logical row height
        self.height = raw.get('height', 0.0)

        # cells in row
        self._cells = Cells(None, self).from_dicts(raw.get('cells', []))


    def __getitem__(self, idx):
        try:
            cell = self._cells[idx]
        except IndexError:
            msg = f'Cell index {idx} out of range'
            raise IndexError(msg)
        else:
            return cell

    def __iter__(self):
        return (cell for cell in self._cells)

    def __len__(self):
        return len(self._cells)


    def append(self, cell:Cell):
        '''Append cell to row and update bbox accordingly.'''
        self._cells.append(cell)


    def store(self) -> dict:
        res = super().store()
        res.update({
            'height': self.height,
            'cells': self._cells.store()
        })

        return res


    def make_docx(self, table, idx_row, border_style:bool):
        '''Create docx table.
            ---
            Args:
              - table: docx table instance
              - idx_row: current row index
              - border_style: whether set border style
        '''  
        # set row height
        docx_row = table.rows[idx_row]

        # to control the layout precisely, set `exact` value, rather than `at least` value
        # the associated steps in MS word: Table Properties -> Row -> Row height -> exactly
        docx_row.height_rule = WD_ROW_HEIGHT.EXACTLY
        
        # NOTE: row height is counted from center-line of top border to center line of bottom border
        docx_row.height = Pt(self.height)

        # set cell style and contents
        for idx_col in range(len(table.columns)):
            self._cells[idx_col].make_docx(table, (idx_row, idx_col), border_style)


# -*- coding: utf-8 -*-

'''
A group of Text/Image or Table block.
@created: 2020-07-22
@author: train8808@gmail.com
'''

from ..common.Collection import Collection
from ..common.base import BlockType
from ..common import utils
from ..common.Block import Block
from ..text.TextBlock import TextBlock
from ..text.ImageBlock import ImageBlock
from ..table.TableBlock import TableBlock
from ..shape.Rectangle import Rectangle
from ..shape import Rectangles # avoid conflits:

class Blocks(Collection):
    '''Block collections.'''
    def __init__(self, instances:list=[], parent=None) -> None:
        ''' A collection of TextBlock and TableBlock instances. 
            ImageBlock is converted to ImageSpan contained in TextBlock.'''
        self._parent = parent # type: Block
        self._instances = []  # type: list[TextBlock or TableBlock]
    
        # Convert all original image blocks to text blocks, i.e. ImageSpan,
        # So we can focus on single TextBlock later on; TableBlock is also combination of TextBlocks.
        for block in instances:
            if isinstance(block, ImageBlock):
                text_block = block.to_text_block()
                self._instances.append(text_block)
            else:
                self._instances.append(block)


    def from_dicts(self, raws:list):
        for raw_block in raws:
            block_type = raw_block.get('type', -1) # type: int
            
            # image block -> text block
            if block_type==BlockType.IMAGE.value:
                block = ImageBlock(raw_block).to_text_block()
            
            # text block
            elif block_type == BlockType.TEXT.value:
                block = TextBlock(raw_block)

            # table block
            elif block_type in (BlockType.EXPLICIT_TABLE.value, BlockType.IMPLICIT_TABLE.value):
                block = TableBlock(raw_block)
            
            else:
                block = None            
            
            # add to list
            self.append(block)
        
        return self


    def text_blocks(self, level=0):
        '''Get text blocks contained in this Collection.
            ---
            Args:
              - level: 
                - 0: text blocks in top level only
                - 1: text blocks deep to table level
        '''
        # top level
        blocks = list(filter(
            lambda block: block.is_text_block(), self._instances))
        
        # table cell level
        if level>0:
            for table in self.table_blocks:
                for row in table:
                    for cell in row:
                        blocks.extend(cell.blocks.text_blocks(level))
        return blocks
    

    def image_spans(self, level=0):
        '''Get ImageSpan contained in this Collection.             
            ---
            Args:
              - level: 
                - 0: image span contained in top level text blocks
                - 1: image span deep to table blocks level

            NOTE:
            No ImageBlock exists in this collection since it's already converted to text block.
        '''
        
        # image span in top text block
        spans = []
        for block in self.text_blocks(level=0):
            spans.extend(block.lines.image_spans)
        
        # image span in table block level
        if level>0:
            for table in self.table_blocks:
                for row in table:
                    for cell in row:
                        spans.extend(cell.blocks.image_spans(level))        
        return spans


    @property
    def explicit_table_blocks(self):
        '''Get explicit table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_explicit_table_block(), self._instances))

    @property
    def implicit_table_blocks(self):
        '''Get implicit table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_implicit_table_block(), self._instances))

    @property
    def table_blocks(self):
        '''Get table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_table_block(), self._instances))


    def clean(self):
        '''Preprocess blocks initialized from the raw layout.'''

        # filter function:
        # - remove negative blocks
        # - remove transformed text: text direction is not (1, 0) or (0, -1)
        # - remove empty blocks
        f = lambda block:   block.is_valid and \
                            block.text.strip() and (
                            block.is_horizontal or block.is_vertical)
        self._instances = list(filter(f, self._instances))
           
        # merge blocks horizontally, e.g. remove overlap blocks, since no floating elements are supported
        # NOTE: It's to merge blocks in physically horizontal direction, i.e. without considering text direction.
        self.sort_in_reading_order().merge_horizontally(text_direction=False)

        return True


    def assign_table_contents(self):
        '''Add Text/Image block lines to associated cells of Table blocks.'''
        # table blocks
        # NOTE: some tables may be already parsed since explicit tables are parsed earlier than implicit tables.
        # It's OK because no text blocks left for parsing such tables at this round.
        tables = self.table_blocks

        if not tables: return

        # collect text blocks in table region        
        blocks_in_tables = [[] for _ in tables] # type: list[list[Block]]
        blocks = []   # type: list[Block]
        for block in self.text_blocks():

            # lines in block for further check if necessary
            lines = block.lines

            # collect blocks contained in table region
            # NOTE: there is a probability that only a part of a text block is contained in table region, 
            # while the rest is in normal block region.
            for table, blocks_in_table in zip(tables, blocks_in_tables):
                # fully contained in one table
                if table.bbox.contains(block.bbox):
                    blocks_in_table.append(block)
                    break
                
                # not possible in current table
                elif not table.bbox.intersects(block.bbox):
                    continue
                
                # deep into line level
                else:
                    text_block = TextBlock()
                    rest_lines = []
                    for line in lines:
                        if table.bbox.intersects(line.bbox):
                            text_block.add(line)
                        else:
                            rest_lines.append(line)
                    
                    # summary
                    blocks_in_table.append(text_block)
                    if rest_lines:
                        lines = rest_lines # for forther check
                    else:
                        break # no more lines
            
            # normal blocks
            else:
                text_block = TextBlock()
                for line in lines:
                    text_block.add(line)
                blocks.append(text_block)

        # assign blocks to associated cells
        # ATTENTION: no nested table is considered
        for table, blocks_in_table in zip(tables, blocks_in_tables):
            # table is parsed already
            if not blocks_in_table: continue
            
            for row in table:
                for cell in row:
                    if not cell: continue
                    # check candidate blocks
                    for block in blocks_in_table: cell.add(block)
                    # merge blocks if contained blocks found
                    cell.blocks.merge_horizontally().split_vertically()

        # sort in natural reading order and update layout blocks
        blocks.extend(tables)
        self.reset(blocks).sort_in_reading_order()


    def collect_table_contents(self):
        ''' Collect bbox, e.g. Line of TextBlock, which may contained in an implicit table region.
            
            Table may exist on the following conditions:
             - (a) lines in blocks are not connected sequently -> determined by current block only
             - (b) multi-blocks are in a same row (horizontally aligned) -> determined by two adjacent blocks
        '''      

        res = [] # type: list[Rectangles.Rectangles]

        table_lines = [] # type: list[Rect]
        new_line = True
        num = len(self._instances)

        for i in range(num):
            block =self._instances[i]
            next_block =self._instances[i+1] if i<num-1 else Block()

            table_end = False
            
            # there is gap between these two criteria, so consider condition (a) only if if it's the first block in new row
            # (a) lines in current block are connected sequently?
            # yes, counted as table lines
            if new_line and block.contains_discrete_lines(): 
                table_lines.extend(block.sub_bboxes)                
                # update line status
                new_line = False            

            # (b) multi-blocks are in a same row: check layout with next block?
            # yes, add both current and next blocks
            if block.horizontally_align_with(next_block):
                # if it's start of new table row: add the first block
                if new_line: 
                    table_lines.extend(block.sub_bboxes)
                
                # add next block
                table_lines.extend(next_block.sub_bboxes)

                # update line status
                new_line = False

            # no, consider to start a new row
            else:
                # table end 
                # - if it's a text line, i.e. no more than one block in a same line
                # - or the next block is also a table
                if new_line or block.is_table_block():
                    table_end = True

                # update line status            
                new_line = True

            # NOTE: close table detecting manually if last block
            if i==num-1:
                table_end = True

            # end of current table
            if table_lines and table_end: 
                # from fitz.Rect to Rectangle type
                rects = [Rectangle().update(line) for line in table_lines]
                res.append(Rectangles.Rectangles(rects))

                # reset table_blocks
                table_lines = []
        
        return res


    def parse_vertical_spacing(self, bbox:tuple):
        ''' Calculate external and internal vertical space for text blocks.
        
            - paragraph spacing is determined by the vertical distance to previous block. 
              For the first block, the reference position is top margin.
            
                It's easy to set before-space or after-space for a paragraph with python-docx,
                so, if current block is a paragraph, set before-space for it; if current block 
                is not a paragraph, e.g. a table, set after-space for previous block (generally, 
                previous block should be a paragraph).
            
            - line spacing is defined as the average line height in current block.

            ---
            Args:
            - bbox: reference boundary of all the blocks
        '''
        if not self._instances: return

        # check text direction
        # normal reading direction by default, i.e. from left to right, 
        # the reference boundary is top border, i.e. bbox[1].
        # regarding vertical text direction, e.g. from bottom to top, left border bbox[0] is the reference
        idx = 1 if self.is_horizontal else 0

        ref_block = self._instances[0]
        ref_pos = bbox[idx]

        for block in self._instances:
            # NOTE: the table bbox is counted on center-line of outer borders, so a half of top border
            # size should be excluded from the calculated vertical spacing
            if block.is_table_block():
                dw = block[0][0].border_width[0] / 2.0 # use top border of the first cell

                # calculate vertical spacing of blocks under this table
                block.parse_vertical_spacing()
            
            else:
                dw = 0.0

            start_pos = block.bbox_raw[idx] - dw
            para_space = max(start_pos-ref_pos, 0.0) # ignore negative value

            # ref to current (paragraph): set before-space for paragraph
            if block.is_text_block():

                # spacing before this paragraph
                block.before_space = start_pos-ref_pos # keep negative value temperally

                # calculate average line spacing in paragraph
                # NOTE: adjust before space if negative value
                block.parse_line_spacing()

            # if ref to current (image): set before-space for paragraph
            elif block.is_image_block():
                block.before_space = para_space

            # ref (paragraph/image) to current: set after-space for ref paragraph        
            elif ref_block.is_text_block() or ref_block.is_image_block():
                ref_block.after_space = para_space

            # situation with very low probability, e.g. ref (table) to current (table)
            # we can't set before space for table in docx, but the tricky way is to
            # create a empty paragraph and set paragraph line spacing and before space
            else:
                block.before_space = para_space

            # update reference block        
            ref_block = block
            ref_pos = ref_block.bbox_raw[idx+2] + dw # assume same bottom border with top one


    def merge_horizontally(self, text_direction=True):
        '''Merge blocks aligned horizontally group by group.
            ---
            Args:
              - text_direction: whether consider text direction.
                if True, detect text direction based on line direction;
                if False, use default direction: from left to right.
        '''
        # get horizontally aligned blocks group by group
        fun = lambda a,b: a.horizontally_align_with(b, factor=0.0, text_direction=text_direction)
        groups = self.group(fun)
        
        # merge blocks in each group
        blocks = []
        for blocks_collection in groups:
            block = blocks_collection._merge_one()
            blocks.append(block)

        self.reset(blocks)

        return self


    def split_vertically(self):
        '''Split block lines in vertical direction.'''
        blocks = [] # type: list[TextBlock]
        for block in self._instances:
            blocks.extend(block.split())
        
        self.reset(blocks)

        return self


    def parse_text_format(self, rects):
        '''Parse text format with style represented by rectangles.
            ---
            Args:
              - rects: Rectangles, potential styles applied on blocks

            NOTE: `parse_text_format` must be implemented by TextBlock, ImageBlock and TableBlock.
        '''
        flag = False
        for block in self._instances:
            if block.parse_text_format(rects):
                flag = True        
        return flag


    def page_margin(self, width:float, height:float):
        '''Calculate page margin.            
            ---
            Args:
            - width: page width
            - height: page height

            Calculation method:
            - left: MIN(bbox[0])
            - right: MIN(left, width-max(bbox[2]))
            - top: MIN(bbox[1])
            - bottom: height-MAX(bbox[3])
        '''
        # return normal page margin if no blocks exist
        if not self._instances:
            return (utils.ITP, ) * 4 # 1 Inch = 72 pt

        # check candidates for left margin:
        list_bbox = list(map(lambda x: x.bbox, self._instances))

        # left margin 
        left = min(map(lambda x: x.x0, list_bbox))

        # right margin
        x_max = max(map(lambda x: x.x1, list_bbox))
        right = width - x_max - utils.DM*10.0  # consider tolerance: leave more free space
        right = min(right, left)     # symmetry margin if necessary
        right = max(right, 0.0)      # avoid negative margin

        # top/bottom margin
        top = min(map(lambda x: x.y0, list_bbox))
        bottom = height-max(map(lambda x: x.y1, list_bbox))
        bottom = max(bottom, 0.0)

        # margin is calculated based on text block only, without considering shape, e.g. table border,
        # so reduce calculated top/bottom margin to left some free space
        top *= 0.5
        bottom *= 0.5

        # use normal margin if calculated margin is large enough
        return (
            min(utils.ITP, left), 
            min(utils.ITP, right), 
            min(utils.ITP, top), 
            min(utils.ITP, bottom)
            )


    def _merge_one(self):
        ''' Merge all text blocks into one text block.
            
            NOTE:            
            Lines in text block must have same property, e.g. height, vertical distance, 
            because average line height is used when create docx. However, the contained lines 
            may be not reasonable after this step. So, this is just a pre-processing step focusing 
            on processing lines in horizontal direction, e.g. merging inline image to its text line.
            A further step, e.g. split lines vertically, must be applied before final making docx.
        '''
        # combine all lines into a TextBlock
        final_block = TextBlock()
        for block in self._instances:
            if not block.is_text_block(): continue
            for line in block.lines:
                final_block.add(line)

        # merge lines/spans contained in this textBlock
        final_block.lines.merge()

        return final_block




# -*- coding: utf-8 -*-

import os
import fitz
from docx import Document

from .layout.Layout import Layout
from .common.base import PlotControl
from .common.pdf import new_page_section


class Converter:
    ''' Read PDF file `pdf_file` with PyMuPDF to get raw layout data page by page, including text, 
        image and the associated properties, e.g. boundary box, font, size, image width, height, 
        then parse it with consideration for docx re-generation structure. Finally, generate docx
        with python-docx.
    '''

    def __init__(self, pdf_file:str, docx_file:str=None, debug:bool=False) -> None:
        ''' Initialize fitz object with given pdf file path; initialize docx object.

            If debug=True, illustration pdf will be created during parsing the raw pdf layout.
        '''
        # pdf/docx filename
        self.filename_pdf = pdf_file
        self.filename_docx = docx_file if docx_file else pdf_file.replace('.pdf', '.docx')
        if os.path.exists(self.filename_docx):
            os.remove(self.filename_docx)

        # fitz object to read pdf
        self._doc_pdf = fitz.open(pdf_file)

        # docx object to write file
        self._doc_docx = Document()

        # layout object: main worker
        self._layout = None # type: Layout

        # fitz object in debug mode: plot page layout
        # file path for this debug pdf: demo.pdf -> debug_demo.pdf
        self.debug_mode = debug
        path, filename = os.path.split(pdf_file)
        self.filename_debug = os.path.join(path, f'debug_{filename}')
        self._doc_debug = fitz.open() if debug else None

        # to serialize layout for debug purpose
        self._filename_debug_layout = os.path.join(path, 'layout.json')


    def __getitem__(self, index):
        if isinstance(index, slice):
            if index.stop is None or index.stop > self._doc_pdf.pageCount:
                stop = self._doc_pdf.pageCount
            else:
                stop = index.stop
            res = [self._doc_pdf[i] for i in range(stop)]
            return res[index]
        else:
            return self._doc_pdf[index]

    def __len__(self):
        return len(self._doc_pdf)

    @property
    def layout(self):
        return self._layout

    @property
    def doc_pdf(self):
        return self._doc_pdf

    @property
    def doc_docx(self):
        return self._doc_docx

    @property
    def _debug_kwargs(self):
        return {
                'debug': self.debug_mode,
                'doc': self._doc_debug,
                'filename': self.filename_debug
            }

    def init(self, page:fitz.Page) -> Layout:
        '''Initialize layout object.'''
        # Layout object based on raw dict
        raw_layout = page.getText('rawdict')
        self._layout = Layout(raw_layout)
        
        # get rectangle shapes from page source
        self._layout.rects.from_stream(self.doc_pdf, page)
        
        # get annotations(comment shapes) from PDF page, e.g. 
        # highlight, underline and strike-through-line        
        self._layout.rects.from_annotations(page)

        # plot raw layout
        if self.debug_mode:
            # new section for current pdf page
            new_page_section(self._doc_debug, self._layout.width, self._layout.height, f'Page {page.number}')

            # initial layout
            self._layout.plot(self._doc_debug, 'Original Text Blocks', key=PlotControl.LAYOUT)
            self._layout.plot(self._doc_debug, 'Original Rectangle Shapes', key=PlotControl.SHAPE)

        return self._layout


    def parse(self, page:fitz.Page):
        '''Parse page layout.'''
        # parse page: text/table/layout format
        self.init(page).parse(**self._debug_kwargs)

        # debug:
        # - save layout plotting as pdf file
        # - write layout information
        if self.debug_mode:
            self._doc_debug.save(self.filename_debug)
            self._layout.serialize(self._filename_debug_layout)

        return self


    def make_page(self):
        '''Create docx page based on parsed layout.'''
        self._layout.make_page(self._doc_docx)
        self.save_docx()


    def extract_tables(self, page:fitz.Page):
        '''Extract table contents.'''
        return self.init(page).extract_tables()


    def save_docx(self):
        '''Save docx file.'''        
        self._doc_docx.save(self.filename_docx)


    def close(self):
        '''Close pdf files.'''
        self._doc_pdf.close()
        if self.debug_mode:
            self._doc_debug.close()


# -*- coding: utf-8 -*-

'''
Parsing TableBlock based on rectangles explicitly or text blocks implicitly.

@created: 2020-08-16
@author: train8808@gmail.com
'''

from ..common.BBox import BBox
from ..common.base import RectType
from ..common import utils
from ..layout.Blocks import Blocks
from ..shape.Rectangle import Rectangle
from ..shape.Rectangles import Rectangles
from ..table.TableBlock import TableBlock
from ..table.Row import Row
from ..table.Cell import Cell


class TablesConstructor:

    def __init__(self, blocks:Blocks, rects:Rectangles):
        '''Object parsing TableBlock.'''
        self._blocks = blocks
        self._rects = rects


    def explicit_tables(self):
        '''parse table structure from rectangle shapes'''
        # group rects: each group may be a potential table
        fun = lambda a,b: a.bbox & b.bbox
        groups = self._rects.group(fun)

        # parse table with each group
        tables = Blocks()
        for group in groups:
            # parse table structure based on rects in border type
            table = self._parse_table_structure(group, detect_border=True)
            tables.append(table)

        # check if any intersection with previously parsed tables
        for group in tables.group(fun):
            # single table
            if len(group)==1:
                table = group[0]
            
            # intersected tables: keep the table with the most cells only 
            # since no floating elements are supported with python-docx
            else:
                sorted_group = sorted(group, 
                            key=lambda table: table.num_rows*table.num_cols)
                table = sorted_group[-1]                    

            # add table to page level
            table.set_explicit_table_block()
            self._blocks.append(table)

        # assign text contents to each table
        self._blocks.assign_table_contents()

        return self._blocks.explicit_table_blocks

    
    def implicit_tables(self, X0:float, X1:float):
        ''' Parse table structure based on the layout of text/image blocks.
            ---
            Args:
            - X0, X1: left and right boundaries of allowed table region

            Since no cell borders exist in this case, there may be various probabilities of table structures. 
            Among which, we use the simplest one, i.e. 1-row and n-column, to make the docx look like pdf.

            Ensure no horizontally aligned blocks in each column, so that these blocks can be converted to
            paragraphs consequently in docx.
        '''    
        if len(self._blocks)<=1: return []      

        # potential bboxes
        tables_bboxes = self._blocks.collect_table_contents()

        # parse tables
        for table_bboxes in tables_bboxes:
            # parse borders based on contents in cell
            table_rects = self._implicit_borders(table_bboxes, X0, X1)

            # get potential cell shading
            table_bbox = table_rects.bbox
            table_rects.extend(filter(
                lambda rect: table_bbox.intersects(rect.bbox), self._rects))

            # parse table: don't have to detect borders since it's determined already
            table = self._parse_table_structure(table_rects, detect_border=False)

            # add parsed table to page level blocks
            # in addition, ignore table if contains only one cell since it's unnecessary for implicit table
            if table and (table.num_rows>1 or table.num_cols>1):
                table.set_implicit_table_block()
                self._blocks.append(table)

        # assign text contents to each table
        self._blocks.assign_table_contents()

        return self._blocks.implicit_table_blocks


    def _parse_table_structure(self, rects:Rectangles, detect_border:bool=True):
        ''' Parse table structure from rects.
            ---
            Args:
            - rects: Rectangles, representing border, shading or text style
            - detect_border: to detect table border if True.

            NOTE: for implicit table, table borders are determined from text blocks in advance,
            so, it's safe to set `detect_border=False`.
        '''

        # --------------------------------------------------
        # mark table borders first
        # --------------------------------------------------
        # exit if no borders exist
        if detect_border and not self._set_table_borders(rects, width_threshold=6.0):
            return None
        
        # --------------------------------------------------
        # group horizontal/vertical borders
        # --------------------------------------------------
        h_borders, v_borders = self._group_explicit_borders(rects)
        if not h_borders or not v_borders:
            # reset borders because it's a invalid table
            self._unset_table_border(rects)
            return None

        # sort
        y_rows = sorted(h_borders)
        x_cols = sorted(v_borders)       
            
        # --------------------------------------------------
        # parse table structure, especially the merged cells
        # -------------------------------------------------- 
        # check merged cells in each row
        merged_cells_rows = []  # type: list[list[int]]
        for i, row in enumerate(y_rows[0:-1]):
            ref_y = (row+y_rows[i+1])/2.0
            ordered_v_borders = [v_borders[k] for k in x_cols]
            row_structure = self._check_merged_cells(ref_y, ordered_v_borders, 'row')
            merged_cells_rows.append(row_structure)

        # check merged cells in each column
        merged_cells_cols = []  # type: list[list[int]]
        for i, col in enumerate(x_cols[0:-1]):
            ref_x = (col+x_cols[i+1])/2.0
            ordered_h_borders = [h_borders[k] for k in y_rows]
            col_structure = self._check_merged_cells(ref_x, ordered_h_borders, 'column')        
            merged_cells_cols.append(col_structure)

        # --------------------------------------------------
        # parse table properties
        # --------------------------------------------------
        table = TableBlock()
        n_rows = len(merged_cells_rows)
        n_cols = len(merged_cells_cols)

        for i in range(n_rows):
            # row object
            row = Row()
            row.height = y_rows[i+1]-y_rows[i]
            
            for j in range(n_cols):
                # if current cell is merged horizontally or vertically, set None.
                # actually, it will be counted in the top-left cell of the merged range.
                if merged_cells_rows[i][j]==0 or merged_cells_cols[j][i]==0:
                    row.append(Cell())
                    continue

                # Now, this is the top-left cell of merged range.
                # A separate cell without merging can also be treated as a merged range 
                # with 1 row and 1 colum, i.e. itself.
                #             
                # check merged columns in horizontal direction
                n_col = 1
                for val in merged_cells_rows[i][j+1:]:
                    if val==0:
                        n_col += 1
                    else:
                        break
                # check merged rows in vertical direction
                n_row = 1
                for val in merged_cells_cols[j][i+1:]:
                    if val==0:
                        n_row += 1
                    else:
                        break

                # cell border rects: merged cells considered
                top = h_borders[y_rows[i]][0]
                bottom = h_borders[y_rows[i+n_row]][0]
                left = v_borders[x_cols[j]][0]
                right = v_borders[x_cols[j+n_col]][0]

                w_top = top.bbox.y1-top.bbox.y0
                w_right = right.bbox.x1-right.bbox.x0
                w_bottom = bottom.bbox.y1-bottom.bbox.y0
                w_left = left.bbox.x1-left.bbox.x0

                # cell bbox
                bbox = (x_cols[j], y_rows[i], x_cols[j+n_col], y_rows[i+n_row])

                # shading rect in this cell
                # modify the cell bbox from border center to inner region
                inner_bbox = (bbox[0]+w_left/2.0, bbox[1]+w_top/2.0, bbox[2]-w_right/2.0, bbox[3]-w_bottom/2.0)
                target_bbox = BBox().update(inner_bbox)
                shading_rect = self._get_rect_with_bbox(target_bbox, rects, threshold=0.9)
                if shading_rect:
                    shading_rect.type = RectType.SHADING # ATTENTION: set shaing type
                    bg_color = shading_rect.color
                else:
                    bg_color = None

                # Cell object
                cell_dict = {
                    'bbox': bbox,
                    'bg_color':  bg_color,
                    'border_color': (top.color, right.color, bottom.color, left.color),
                    'border_width': (w_top, w_right, w_bottom, w_left),
                    'merged_cells': (n_row, n_col),
                }
                row.append(Cell(cell_dict))
                    
            # one row finished
            # check table: the first cell in first row MUST NOT be None
            if i==0 and not row[0]:
                # reset borders because it's a invalid table
                self._unset_table_border(rects)
                return None

            table.append(row)

        return table


    @staticmethod
    def _set_table_borders(rects:Rectangles, width_threshold:float=6.0):
        ''' Detect table borders from rects.
            ---
            Args:
            - rects: all rects in potential table region
            - width_threshold: float, suppose border width is lower than this threshold value

            Cell borders are detected based on the experiences that:
              - compared to cell shading, the size of cell border never exceeds 6 pt
              - compared to text format, cell border always has intersection with other rects

            NOTE: cell shading is determined after the table structure is parsed from these cell borders.
        '''
        # Get all rects with on condition: size < 6 Pt
        thin_rects = [] # type: list[Rectangle]
        for rect in rects:
            x0, y0, x1, y1 = rect.bbox_raw
            if min(x1-x0, y1-y0) <= width_threshold:
                thin_rects.append(rect)

        # These thin rects may be cell borders, or text format, e.g. underline within cell.
        # Compared to text format, cell border always has intersection with other rects
        borders = [] # type: list[Rectangle]
        for rect in thin_rects:
            # check intersections with other rect
            for other_rect in thin_rects:
                if rect==other_rect: continue
                # it's a cell border if intersection found
                # Note: if the intersection is an edge, method `intersects` returns False, while
                # the operator `&` return True. So, `&` is used here.
                if rect.bbox & other_rect.bbox: 
                    borders.append(rect)
                    break

        # at least two inner borders exist for a normal table
        if len(borders)>=2:
            # set table border type
            for rect in borders:
                rect.type = RectType.BORDER
            return True
        else:
            return False


    @staticmethod
    def _unset_table_border(rects:Rectangles):
        '''Unset table border type.'''
        for rect in rects:
            if rect.type==RectType.BORDER:
                rect.type = RectType.UNDEFINED


    def _group_explicit_borders(self, rects:Rectangles):
        ''' Collect explicit borders in horizontal and vertical groups respectively.'''        

        h_borders = {} # type: dict [float, Rectangles]
        v_borders = {} # type: dict [float, Rectangles]
        h_outer = []   # type: list[float]
        v_outer = []   # type: list[float]

        for rect in rects.border_rects:
            # group horizontal borders in each row
            if rect.bbox.width > rect.bbox.height:
                # row centerline
                y = round((rect.bbox.y0 + rect.bbox.y1) / 2.0, 1)

                # ignore minor error resulting from different border width
                for y_ in h_borders:
                    if abs(y-y_)<utils.DM:
                        h_borders[y_].append(rect)
                        break
                else:
                    h_borders[y] = Rectangles([rect])
                
                # candidates for vertical outer border
                v_outer.extend([rect.bbox.x0, rect.bbox.x1])

            # group vertical borders in each column
            else:
                # column centerline
                x = round((rect.bbox.x0 + rect.bbox.x1) / 2.0, 1)
                
                # ignore minor error resulting from different border width
                for x_ in v_borders:
                    if abs(x-x_)<utils.DM:
                        v_borders[x_].append(rect)
                        break
                else:
                    v_borders[x] = Rectangles([rect])
                
                # candidates for horizontal outer border
                h_outer.extend([rect.bbox.y0, rect.bbox.y1])

        # at least 2 inner borders exist
        if not h_borders or not v_borders:
            return None, None

        # Note: add dummy borders if no outer borders exist
        # check whether outer borders exists in collected borders
        if h_borders:
            top_rects = h_borders[min(h_borders)]
            bottom_rects = h_borders[max(h_borders)]
            left   = min(v_outer)
            right  = max(v_outer)
        else:
            top_rects = Rectangles()
            bottom_rects = Rectangles()
            left   = None
            right  = None

        if v_borders:
            left_rects = v_borders[min(v_borders)]
            right_rects = v_borders[max(v_borders)]
            top   = min(h_outer)
            bottom  = max(h_outer)
        else:
            left_rects = Rectangles()
            right_rects = Rectangles()
            top   = None
            bottom  = None    

        c = utils.RGB_value((1,1,1))
        if not self._exist_outer_border(top_rects, top, 'h'):
            h_borders[top] = Rectangles([
                Rectangle({
                    'bbox': (left, top, right, top),
                    'color': c
                })])
        if not self._exist_outer_border(bottom_rects, bottom, 'h'):
            h_borders[bottom] = Rectangles([
                Rectangle({
                    'bbox': (left, bottom, right, bottom),
                    'color': c
                })])
        if not self._exist_outer_border(left_rects, left, 'v'):
            v_borders[left] = Rectangles([
                Rectangle({
                    'bbox': (left, top, left, bottom),
                    'color': c
                })])
        if not self._exist_outer_border(right_rects, right, 'v'):
            v_borders[right] = Rectangles([
                Rectangle({
                    'bbox': (right, top, right, bottom),
                    'color': c
                })])

        return h_borders, v_borders


    @staticmethod
    def _exist_outer_border(rects:Rectangles, target:float, direction:str='h') -> bool:
        ''' Check outer borders: whether target border exists in collected borders.
            ---
            Args:
            - rects: all rects in potential table region
            - target: float, target position of outer border
            - direction: str, 'h'->horizontal border; 'v'->vertical border
        '''
        # no target outer border needed
        if target==None: return True

        # need outer border if no borders exist
        if not rects: return False
        
        # considering direction
        idx = 1 if direction=='h' else 0
        
        # centerline of source borders
        source = round((rects[0].bbox_raw[idx+2] + rects[0].bbox_raw[idx]) / 2.0, 1)
        # max width of source borders
        width = max(map(lambda rect: rect.bbox_raw[idx+2]-rect.bbox_raw[idx], rects))

        target = round(target, 1)
        width = round(width, 1)

        return abs(target-source) <= width


    @staticmethod
    def _check_merged_cells(ref:float, borders:list, direction:str='row'):
        ''' Check merged cells in a row/column. 
            ---
            Args:
              - ref: y (or x) coordinate of horizontal (or vertical) passing-through line
              - borders: list[Rectangles], a list of vertical (or horizontal) rects list in a column (or row)
              - direction: 'row' - check merged cells in row; 'column' - check merged cells in a column

            Taking cells in a row (direction=0) for example, give a horizontal line (y=ref) passing through this row, 
            check the intersection with vertical borders. The n-th cell is merged if no intersection with the n-th border.
            
        '''
        res = [] # type: list[int]
        for rects in borders[0:-1]:
            # multi-lines exist in a row/column
            for rect in rects:

                # reference coordinates depending on checking direction
                if direction=='row':
                    _, ref0, _, ref1 = rect.bbox_raw
                else:
                    ref0, _, ref1, _ = rect.bbox_raw

                # 1) intersection found
                if ref0 < ref < ref1:
                    res.append(1)
                    break
                
                # 2) reference line locates below current rect:
                # still have a chance to find intersection with next rect, but,
                # no chance if this is the last rect, see the else-clause
                elif ref > ref1:
                    continue

                # 3) current rect locates below the reference line:
                # no intersection is possible any more
                elif ref < ref0:
                    res.append(0)
                    break
            
            # see notes 2), no change any more
            else:
                res.append(0)

        return res


    @staticmethod
    def _get_rect_with_bbox(target:BBox, rects:Rectangles, threshold:float):
        '''Get rect from rects on condition that contained in given bbox.'''
        s = target.bbox.getArea()
        if not s: return None

        for rect in rects:
            intersection = target.bbox & rect.bbox
            if intersection.getArea() / s >= threshold:
                res = rect
                break
        else:
            res = None
        return res


    def _implicit_borders(self, rects:Rectangles, X0:float, X1:float):
        ''' Parsing borders based on rects representing contents in table cells.
            ---
            Args:
            - rects: Rectangles, representing contents in table cells
            - X0, X1: default left and right outer borders
        '''
        # boundary box (considering margin) of all line box
        margin = 1
        x0 = X0 - margin
        y0 = min([rect.bbox.y0 for rect in rects]) - margin
        x1 = X1 + margin
        y1 = max([rect.bbox.y1 for rect in rects]) + margin
        border_bbox = (x0, y0, x1, y1)

        # centerline of outer borders
        borders = [
            (x0, y0, x1, y0), # top
            (x1, y0, x1, y1), # right
            (x0, y1, x1, y1), # bottom
            (x0, y0, x0, y1)  # left
        ]

        # centerline of inner borders
        inner_borders = self._borders_from_bboxes(rects, border_bbox)
        borders.extend(inner_borders)
        
        # all centerlines to rectangle shapes
        res = Rectangles()
        color = utils.RGB_value((1,1,1))
        for border in borders: 
            # set an non-zero width for border check; won't draw border in docx for implicit table
            bbox = utils.expand_centerline(border[0:2], border[2:], width=0.2) 
            if not bbox: continue

            # create Rectangle object and set border style
            rect = Rectangle({
                'bbox': bbox,
                'color': color
            })
            rect.type = RectType.BORDER
            res.append(rect)

        return res


    def _borders_from_bboxes(self, rects:Rectangles, border_bbox:tuple):
        ''' Calculate the surrounding borders of given bbox-es.
            ---
            Args:
            - rects: contents bbox-es in table cells
            - border_bbox: boundary box of table region
            
            These borders construct table cells. Considering the re-building of cell content in docx, 
            - only one bbox is allowed in a line;
            - but multi-lines are allowed in a cell.
        '''
        borders = []  # type: list[tuple[float]]

        # collect bbox-ex column by column
        X0, Y0, X1, Y1 = border_bbox
        cols_rects = self._column_borders_from_bboxes(rects)
        col_num = len(cols_rects)

        for i in range(col_num):
            # add column border
            x0 = X0 if i==0 else (cols_rects[i-1].bbox.x1 + cols_rects[i].bbox.x0) / 2.0
            x1 = X1 if i==col_num-1 else (cols_rects[i].bbox.x1 + cols_rects[i+1].bbox.x0) / 2.0

            if i<col_num-1:
                borders.append((x1, Y0, x1, Y1))

            # collect bboxes row by row        
            rows_rects = self._row_borders_from_bboxes(cols_rects[i])

            # NOTE: unnecessary to split row if the count of row is 1
            row_num = len(rows_rects)
            if row_num==1: continue
        
            for j in range(row_num):
                # add row border
                y0 = Y0 if j==0 else (rows_rects[j-1].bbox.y1 + rows_rects[j].bbox.y0) / 2.0
                y1 = Y1 if j==row_num-1 else (rows_rects[j].bbox.y1 + rows_rects[j+1].bbox.y0) / 2.0
                
                # it's Ok if single bbox in a line
                # if len(rows_rects[j])<2:
                #     continue

                # otherwise, add row border and check borders further
                if j==0:
                    borders.append((x0, y1, x1, y1))
                elif j==row_num-1:
                    borders.append((x0, y0, x1, y0))
                else:
                    borders.append((x0, y0, x1, y0))
                    borders.append((x0, y1, x1, y1))

                # recursion
                _borders = self._borders_from_bboxes(rows_rects[j], (x0, y0, x1, y1))
                borders.extend(_borders)        

        return borders


    @staticmethod
    def _column_borders_from_bboxes(rects:Rectangles):
        ''' split bbox-es into column groups and add border for adjacent two columns.'''
        # sort bbox-ex in column first: from left to right, from top to bottom
        rects.sort_in_line_order()
        
        #  bboxes list in each column
        cols_rects = [] # type: list[Rectangles]

        # collect bbox-es column by column
        col_rect = Rectangle()
        for rect in rects:
            # same column group if vertically aligned
            if col_rect.vertically_align_with(rect):
                cols_rects[-1].append(rect)
            
            # otherwise, start a new column group
            else:
                cols_rects.append(Rectangles([rect]))
                col_rect = Rectangle() # reset
                
            col_rect.union(rect)

        return cols_rects


    @staticmethod
    def _row_borders_from_bboxes(rects:Rectangles):
        ''' split bbox-es into row groups and add border for adjacent two rows.'''
        # sort bbox-ex in row first mode: from top to bottom, from left to right
        rects.sort_in_reading_order()

        #  bboxes list in each row
        rows_rects = [] # type: list[Rectangles]

        # collect bbox-es row by row
        row_rect = Rectangle()
        for rect in rects:
            # same row group if horizontally aligned
            if row_rect.horizontally_align_with(rect):
                rows_rects[-1].append(rect)
            
            # otherwise, start a new row group
            else:
                rows_rects.append(Rectangles([rect]))
                row_rect = Rectangle() # reset

            row_rect.union(rect)

        return rows_rects


# -*- coding: utf-8 -*-

'''
A group of instances, e.g. instances, Spans, Rectangles.

@created: 2020-07-24
@author: train8808@gmail.com
'''

from .BBox import BBox
from .Block import Block
from .base import IText, TextDirection

class Collection(IText):
    '''Collection of specific instances.'''
    def __init__(self, instances:list=[], parent=None) -> None:
        ''' Construct text line from a list of raw line dict.'''
        self._instances = instances if instances else [] # type: list[BBox]
        self._parent = parent # type: Block


    def __getitem__(self, idx):
        try:
            instances = self._instances[idx]
        except IndexError:
            msg = f'Collection index {idx} out of range'
            raise IndexError(msg)
        else:
            return instances


    def __iter__(self):
        return (instance for instance in self._instances)


    def __len__(self):
        return len(self._instances)


    @property
    def bbox(self):
        '''bbox of combined collection.'''
        Box = BBox()
        for instance in self._instances:
            Box.union(instance)
        return Box.bbox


    @property
    def text_direction(self):
        '''Get text direction. All instances must have same text direction.''' 
        if self._instances and hasattr(self._instances[0], 'text_direction'):
            res = set(instance.text_direction for instance in self._instances)
            if len(res)==1:
                return list(res)[0]

        # normal direction by default
        return TextDirection.LEFT_RIGHT 


    def from_dicts(self, *args, **kwargs):
        '''Construct Collection from a list of dict.'''
        raise NotImplementedError


    def append(self, bbox:BBox):
        '''Append an instance and update parent's bbox accordingly.'''
        if not bbox: return
        self._instances.append(bbox)
        if not self._parent is None: # Note: `if self._parent` does not work here
            self._parent.union(bbox)


    def extend(self, bboxes:list):
        '''Append a list of instances.'''
        for bbox in bboxes:
            self.append(bbox)


    def insert(self, nth:int, bbox:BBox):
        '''Insert a BBox and update parent's bbox accordingly.'''
        if not bbox: return
        self._instances.insert(nth, bbox)
        if not self._parent is None:
            self._parent.union(bbox)


    def sort_in_reading_order(self):
        '''Sort collection instances in reading order (considering text direction), e.g.
            for normal reading direction: from top to bottom, from left to right.
        '''
        if self.is_horizontal:
            self._instances.sort(key=lambda instance: (instance.bbox.y0, instance.bbox.x0, instance.bbox.x1))
        else:
            self._instances.sort(key=lambda instance: (instance.bbox.x0, instance.bbox.y1, instance.bbox.y0))
        return self


    def sort_in_line_order(self):
        '''Sort collection instances in a physical with text direction considered, e.g.
            for normal reading direction: from left to right.
        '''
        if self.is_horizontal:
            self._instances.sort(key=lambda instance: (instance.bbox.x0, instance.bbox.y0, instance.bbox.x1))
        else:
            self._instances.sort(key=lambda instance: (instance.bbox.y1, instance.bbox.x0, instance.bbox.y0))
        return self


    def reset(self, bboxes:list=[]):
        '''Reset instances list.'''
        self._instances = []
        self.extend(bboxes)
        return self


    def store(self) -> list:
        '''Store attributes in json format.'''
        return [ instance.store() for instance in self._instances ]


    def group(self, fun):
        '''group instances according to user defined criterion.
            ---
            Args:
              - fun: function with 2 parameters (BBox) representing 2 instances, and return bool
            
            Examples:
            ```
            # group instances intersected with each other
            fun = lambda a,b: a & b
            # group instances aligned horizontally
            fun = lambda a,b: a.horizontally_aligned_with(b)
            ```
        '''
        groups = [] # type: list[Collection]
        counted_index = set() # type: set[int]

        # check each instance to the others
        for i,instance in enumerate(self._instances):

            # do nothing if current rect has been considered already
            if i in counted_index:
                continue

            # start a new group
            group = { i }

            # get intersected instances
            self._group_instances(instance, group, fun)

            # update counted instances
            counted_index = counted_index | group

            # add rect to groups
            group_instances = [self._instances[x] for x in group]
            instances = self.__class__(group_instances)
            groups.append(instances)

        return groups


    def _group_instances(self, bbox:BBox, group:set, fun):
        ''' Get instances related to given bbox.
            ---
            Args:
              - bbox: reference bbox
              - group: set[int], a set() of index of intersected instances
              - fun: define the relationship with reference bbox. 
                    2 parameters: instance 1 and instance 2, return bool.
        '''

        idx = 1 if self.is_horizontal else 0

        for i, target in enumerate(self._instances):

            # ignore bbox already processed
            if i in group: continue

            # # ignore instances with different text direction
            # if self.text_direction != bbox.text_direction:
            #     continue

            # if satisfying given relationship, check bboxs further
            if fun(bbox, target):
                group.add(i)
                self._group_instances(target, group, fun)

            # it's sorted already, so no relationship exists if not intersected in vertical direction 
            else:
                if target.bbox_raw[idx] > bbox.bbox_raw[idx+2]:
                    break


#!/usr/bin/python3
# -*- coding: utf-8 -*-


from pdf2docx.converter import Converter


def parse(pdf_file, docx_file, start=0, end=None, pages=[], debug=False):
    ''' Run the pdf2docx parser.
    
        Args:
            pdf_file (str) : PDF filename to read from
            docx_file (str): DOCX filename to write to
            start (int)    : first page to process, starting from zero
            end (int)      : last page to process, starting from zero
            pages (list)   : range of pages
            debug          : create illustration pdf showing layouts if True, else do nothing
    '''

    cv = Converter(pdf_file, docx_file, debug)

    # parsing arguments
    pdf_len = len(cv)
    if pages: 
        pdf_pages = [cv[int(x)] for x in pages]
    else:
        end = end or pdf_len
        pdf_pages = cv[int(start):int(end)]

    # process page by page
    for page in pdf_pages:
        print(f"Processing {page.number}/{pdf_len-1}...")
        cv.parse(page).make_page()

    # close pdf
    cv.close()
    

def extract_tables(pdf_file, start=0, end=None, pages=[]):
    ''' Extract table content from pdf pages.
    
        Args:
            pdf_file (str) : PDF filename to read from
            start (int)    : first page to process, starting from zero
            end (int)      : last page to process, starting from zero
            pages (list)   : range of pages
    '''

    cv = Converter(pdf_file)

    # parsing arguments
    pdf_len = len(cv)
    if pages: 
        pdf_pages = [cv[int(x)] for x in pages]
    else:
        end = end or pdf_len
        pdf_pages = cv[int(start):int(end)]

    # process page by page
    tables = []
    for page in pdf_pages:
        print(f"Processing {page.number}/{pdf_len-1}...")
        page_tables = cv.extract_tables(page)
        tables.extend(page_tables)

    cv.close()

    return tables


def main():
    import fire
    fire.Fire(parse)


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-

'''
A group of Rectangle instances focusing on table parsing process.

@created: 2020-07-22
@author: train8808@gmail.com
'''

from .Rectangle import Rectangle
from ..common.base import RectType
from ..common.Collection import Collection
from ..common import utils
from ..common import pdf


class Rectangles(Collection):

    @property
    def border_rects(self):
        '''Rectangles in border type.'''
        return list(filter(
            lambda rect: rect.type==RectType.BORDER, self._instances))

    def from_annotations(self, page):
        ''' Get shapes, e.g. Line, Square, Highlight, from annotations(comment shapes) in PDF page.
            ---
            Args:
            - page: fitz.Page, current page
        '''
        rects = pdf.rects_from_annotations(page)
        for rect in rects:
            self._instances.append(Rectangle(rect))

        return self


    def from_stream(self, doc, page):
        ''' Get rectangle shapes, e.g. highlight, underline, table borders, from page source contents.
            ---
            Args:
            - doc: fitz.Document representing the pdf file
            - page: fitz.Page, current page
        '''
        rects = pdf.rects_from_stream(doc, page)
        for rect in rects:
            self._instances.append(Rectangle(rect))

        return self


    def clean(self):
        '''Clean rectangles:
            - delete rectangles fully contained in another one (beside, they have same bg-color)
            - join intersected and horizontally aligned rectangles with same height and bg-color
            - join intersected and vertically aligned rectangles with same width and bg-color
        '''
        # sort in reading order
        self.sort_in_reading_order()

        # skip rectangles with both of the following two conditions satisfied:
        #  - fully or almost contained in another rectangle
        #  - same filling color with the containing rectangle
        rects_unique = [] # type: list [Rectangle]
        rect_changed = False
        for rect in self._instances:
            for ref_rect in rects_unique:
                # Do nothing if these two rects in different bg-color
                if ref_rect.color!=rect.color: continue     

                # combine two rects in a same row if any intersection exists
                # ideally the aligning threshold should be 1.0, but use 0.98 here to consider tolerance
                if rect.horizontally_align_with(ref_rect, 0.98): 
                    main_bbox = utils.get_main_bbox(rect.bbox, ref_rect.bbox, 0.0)

                # combine two rects in a same column if any intersection exists
                elif rect.vertically_align_with(ref_rect, 0.98):
                    main_bbox = utils.get_main_bbox(rect.bbox, ref_rect.bbox, 0.0)

                # combine two rects if they have a large intersection
                else:
                    main_bbox = utils.get_main_bbox(rect.bbox, ref_rect.bbox, 0.5)

                if main_bbox:
                    rect_changed = True
                    ref_rect.update(main_bbox)
                    break            
            else:
                rects_unique.append(rect)
                
        # update layout
        if rect_changed:
            self._instances = rects_unique

        return rect_changed


# -*- coding: utf-8 -*-

'''
Base class for text/image/table blocks.

@created: 2020-07-23
@author: train8808@gmail.com
'''

from .base import BlockType
from .BBox import BBox


class Block(BBox):
    '''Text block.'''
    def __init__(self, raw:dict={}):
        super(Block, self).__init__(raw)
        self._type = BlockType.UNDEFINED

        # spacing attributes
        self.before_space = raw.get('before_space', 0.0)
        self.after_space = raw.get('after_space', 0.0)
        self.line_space = raw.get('line_space', 0.0)


    @property
    def sub_bboxes(self):
        '''sub-region bbox of this block, e.g. Lines in TextBlock. Return self.bbox by default.'''
        return [self.bbox]

    def is_text_block(self):
        return self._type==BlockType.TEXT

    def is_image_block(self):
        return self._type==BlockType.IMAGE

    def is_explicit_table_block(self):
        return self._type==BlockType.EXPLICIT_TABLE

    def is_implicit_table_block(self):
        return self._type==BlockType.IMPLICIT_TABLE

    def is_table_block(self):
        return self.is_explicit_table_block() or self.is_implicit_table_block()

    def set_text_block(self):
        self._type = BlockType.TEXT

    def set_image_block(self):
        self._type = BlockType.IMAGE

    def set_explicit_table_block(self):
        self._type = BlockType.EXPLICIT_TABLE

    def set_implicit_table_block(self):
        self._type = BlockType.IMPLICIT_TABLE

    def compare(self, block, threshold:float=0.9):
        '''whether has same bbox and vertical spacing with given block.
            ---
            Args:
              - block: instance to compare
              - threshold: two bboxes are considered same if the overlap area exceeds threshold.

            NOTE: the vertical spacing has most important impacts on the layout of converted docx.
        '''
        res, msg = super().compare(block, threshold)
        if not res:
            return res, msg
        
        if self.before_space != block.before_space:
            return False, f'Inconsistent before space @ {self.bbox_raw}:\n{self.before_space} v.s. {block.before_space}'

        if self.after_space != block.after_space:
            return False, f'Inconsistent after space @ {self.bbox_raw}:\n{self.after_space} v.s. {block.after_space}'

        if self.line_space != block.line_space:
            return False, f'Inconsistent line space @ {self.bbox_raw}:\n{self.line_space} v.s. {block.line_space}'

        return True, ''
        

    def store(self):
        '''Store attributes in json format.'''
        res = super().store()
        res.update({
            'type': self._type.value,
            'before_space': self.before_space,
            'after_space': self.after_space,
            'line_space': self.line_space
            })
        return res

    def contains_discrete_lines(self, distance:float=25, threshold:int=3):
        ''' Check whether lines in block are discrete, False by default. 
            Rewrite it if necessary, e.g. in TextBlock.
        '''
        return False


    def plot(self, *args, **kwargs):
        '''Plot block bbox in PDF page.'''
        raise NotImplementedError


    def parse_text_format(self, *args, **kwargs):
        '''Parse text format.'''
        raise NotImplementedError


    def make_docx(self, *args, **kwargs):
        '''Create associated docx element, e.g. TextBlock/ImageBlock -> paragraph.'''
        raise NotImplementedError


# -*- coding: utf-8 -*-

'''
Table block object parsed from raw image and text blocks.

@created: 2020-07-22
@author: train8808@gmail.com
---

{
    'type': int
    'bbox': (x0, y0, x1, y1),
    'rows': [
        {
            "bbox": (x0, y0, x1, y1),
            "height": float,
            "cells": [
                {
                    'bbox': (x0, y0, x1, y1),
                    'border_color': (sRGB,,,), # top, right, bottom, left
                    'bg_color': sRGB,
                    'border_width': (,,,),
                    'merged_cells': (x,y), # this is the bottom-right cell of merged region: x rows, y cols
                    'blocks': [ {text blocks} ]
                }, # end of cell
                {},
                None, # merged cell
                ...
            ]
        }, # end of row
        {...} # more rows
    ] # end of row
}
'''


from .Row import Row
from .Rows import Rows
from ..common.Block import Block
from ..common import docx


class TableBlock(Block):
    '''Text block.'''
    def __init__(self, raw:dict={}):
        super(TableBlock, self).__init__(raw)

        # collect rows
        self._rows = Rows(None, self).from_dicts(raw.get('rows', []))

        # explicit table by default
        self.set_explicit_table_block()

    def __getitem__(self, idx):
        try:
            row = self._rows[idx]
        except IndexError:
            msg = f'Row index {idx} out of range'
            raise IndexError(msg)
        else:
            return row

    def __iter__(self):
        return (row for row in self._rows)

    def __len__(self):
        return len(self._rows)

    @property
    def num_rows(self):
        return len(self._rows)

    @property
    def num_cols(self):
        return len(self._rows[0]) if self.num_rows else 0

    @property
    def text(self):
        '''Get text contained in each cell.'''
        return [ [cell.text for cell in row] for row in self._rows ]

    
    def append(self, row:Row):
        '''Append row to table and update bbox accordingly.'''
        self._rows.append(row)


    def store(self):
        res = super().store()
        res.update({
            'rows': self._rows.store()
        })
        return res


    def plot(self, page, content=True, style=False):
        '''Plot table block, i.e. cell/line/span.
            ---
            Args:
              - page: fitz.Page object
              - style: plot cell style if True, e.g. border width, shading
              - content: plot text blocks if True
        '''
        for row in self._rows:
            for cell in row:
                # ignore merged cells
                if not cell: continue            
                
                # plot different border colors for explicit / implicit tables when style=False, 
                # i.e. table illustration, rather than real style of explicit table
                bc = (1,0,0) if self.is_explicit_table_block() else (0.6,0.7,0.8)
                cell.plot(page, content=content, style=style, color=bc)

    
    def parse_text_format(self, rects):
        '''Parse text format for blocks contained in each cell.
            ---
            Args:
              - rects: Rectangles, format styles are represented by these rectangles.
        '''
        for row in self._rows:
            for cell in row:
                if not cell: continue
                cell.blocks.parse_text_format(rects)
        
        return True # always return True if table is parsed


    def parse_vertical_spacing(self):
        ''' Calculate vertical space for blocks contained in table cells.'''
        for row in self._rows:
            for cell in row:
                if not cell: continue

                # reference for vertical spacing is dependent on text direction
                x0,y0,x1,y1 = cell.bbox_raw
                w_top, w_right, w_bottom, w_left = cell.border_width
                bbox = (x0+w_left/2.0, y0+w_top/2.0, x1-w_right/2.0, y1-w_bottom/2.0)
                cell.blocks.parse_vertical_spacing(bbox)


    def make_docx(self, table, page_margin:tuple):
        '''Create docx table.
            ---
            Args:
              - table: docx table instance
              - page_margin: page margin (left, right, top, bottom)
        '''
        # set indent
        left, *_ = page_margin
        pos = self.bbox.x0-left
        docx.indent_table(table, pos)

        # set format and contents row by row
        border_style = self.is_explicit_table_block() # border style for explicit table only
        for idx_row in range(len(table.rows)):
            self._rows[idx_row].make_docx(table, idx_row, border_style)